from importlib.resources import files
from pathlib import Path

"""Constants in the seedcase_sprout module."""

"""The format of the timestamp used in batch file names."""
BATCH_TIMESTAMP_FORMAT = "%Y-%m-%dT%H%M%SZ"

"""Regex pattern for timestamps with the format '%Y-%m-%dT%H%M%SZ'. Must match the
format used by BATCH_TIMESTAMP_FORMAT"""
BATCH_TIMESTAMP_PATTERN = r"\d{4}-\d{2}-\d{2}T\d{6}Z"

"""The name of the timestamp column added to the batch data (only used internally)."""
BATCH_TIMESTAMP_COLUMN_NAME = "_batch_file_timestamp_"

TEMPLATES_PATH = Path(str(files("seedcase_sprout").joinpath("templates")))



from seedcase_sprout.check_datapackage import CheckError
from seedcase_sprout.sprout_checks.exclude_non_sprout_resource_errors import (
    exclude_non_sprout_resource_errors,
)


def test_returns_unaltered_empty_list():
    """Should not alter an empty list."""
    assert exclude_non_sprout_resource_errors([]) == []


def test_returns_only_sprout_related_errors():
    """Should only remove errors not relevant for Sprout."""
    errors = [
        CheckError(
            message="'data' is a required property",
            json_path="$.data",
            validator="required",
        ),
        CheckError(
            message="'name' is a required property",
            json_path="$.name",
            validator="required",
        ),
        CheckError(
            message="123 is not of type 'array'", json_path="$.path", validator="type"
        ),
        CheckError(
            message="123 is not of type 'string'", json_path="$.path", validator="type"
        ),
        CheckError(
            message="123 is not of type 'array'",
            json_path="$.sources",
            validator="type",
        ),
    ]

    assert exclude_non_sprout_resource_errors(errors) == [
        CheckError(
            message="'name' is a required property",
            json_path="$.name",
            validator="required",
        ),
        CheckError(
            message="123 is not of type 'string'", json_path="$.path", validator="type"
        ),
        CheckError(
            message="123 is not of type 'array'",
            json_path="$.sources",
            validator="type",
        ),
    ]



import polars as pl
from pytest import raises

from seedcase_sprout.examples import (
    example_data_all_polars_types,
    example_resource_properties_all_polars_types,
)
from seedcase_sprout.extract_field_properties import (
    extract_field_properties,
)
from seedcase_sprout.properties import FieldProperties, ResourceProperties


def _keep_extractable_properties(
    example_properties: ResourceProperties,
) -> list[FieldProperties]:
    """Filter example properties to only keep the extractable properties."""
    assert example_properties.schema
    assert example_properties.schema.fields
    return list(
        map(
            lambda field: FieldProperties(name=field.name, type=field.type),
            example_properties.schema.fields,
        )
    )


def test_properties_are_extracted_correctly():
    """Test that the resource properties are extracted correctly from the data."""
    # Given, when
    extracted_field_properties = extract_field_properties(
        example_data_all_polars_types()
    )
    expected_field_properties = _keep_extractable_properties(
        example_resource_properties_all_polars_types()
    )
    # Then
    assert extracted_field_properties == expected_field_properties


def test_throw_error_with_empty_data():
    """Test that an error is thrown when the data is empty."""
    with raises(ValueError):
        extract_field_properties(pl.DataFrame([]))



from pathlib import Path

from seedcase_sprout import (
    example_package_properties,
    write_properties,
)


def create_test_data_package(tmp_path: Path) -> Path:
    """Creates a package file structure (with empty files) for path function tests.

    Args:
        tmp_path: Path to a temporary folder.

    Returns:
        Path of package.
    """
    tmp_path.mkdir(parents=True, exist_ok=True)
    write_properties(
        properties=example_package_properties(), path=tmp_path / "datapackage.json"
    )

    (tmp_path / "README.md").touch()

    return tmp_path



import polars as pl

from seedcase_sprout.check_data import check_data
from seedcase_sprout.check_properties import (
    check_resource_properties,
)
from seedcase_sprout.constants import BATCH_TIMESTAMP_COLUMN_NAME
from seedcase_sprout.internals import _get_nested_attr
from seedcase_sprout.properties import ResourceProperties


def join_resource_batches(
    data_list: list[pl.DataFrame], resource_properties: ResourceProperties
) -> pl.DataFrame:
    """Join all the batch resource DataFrames into a single (Polars) DataFrame.

    This function takes a list of DataFrames, joins them together and drops any
    duplicate observational units based on the primary key from `resource_properties`.
    Then, it confirms that the data are correct against the `resource_properties` after
    the join.

    The observational unit is the primary key of the resource. For example, if a person
    is part of a research study and has multiple observations, the person's ID and the
    date of collection would be the observational unit.

    If there are any duplicate observational units in the data, only the most recent
    observational unit will be kept based on the timestamp of the batch file. This way,
    if there are any errors or mistakes in older batch files that have been corrected in
    later files, the mistake will be kept in the batch file, but won't be included in
    the `data.parquet` file.

    Args:
        data_list: A list of Polars DataFrames for all the batch files. Use
            `read_resource_batches()` to get a list of DataFrames that have been
            checked against the properties individually.
        resource_properties: The `ResourceProperties` object that contains the
            properties of the resource to check the data against.

    Returns:
        A single DataFrame object of all the batch data with duplicate
            observational units removed.

    Raises:
        ValueError: If an empty `data_list` is provided.
        polars.exceptions.ShapeError: If the dataframes in data_list have different
            shapes, such as mismatched column names or numbers.
        polars.exceptions.SchemaError: If the dataframes in data_list have
            different schemas, e.g., their column data types don't match.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            resource_properties = sp.example_resource_properties()
            sp.write_resource_batch(sp.example_data(), resource_properties)
            batches = sp.read_resource_batches(resource_properties=resource_properties)

            sp.join_resource_batches(batches, resource_properties)
        ```
    """
    check_resource_properties(resource_properties)

    if data_list == []:
        raise ValueError(
            "Could not join resource batches because an empty `data_list` was "
            f"provided. The batch folder for the resource '{resource_properties.name}' "
            "may be empty."
        )

    data = pl.concat(data_list)
    primary_key = _get_nested_attr(resource_properties, "schema.primary_key")
    data = _drop_duplicate_obs_units(data, primary_key)

    check_data(data, resource_properties)

    return data


def _drop_duplicate_obs_units(
    data: pl.DataFrame, primary_key: list[str] | str | None
) -> pl.DataFrame:
    """Drop duplicates based on the primary key and keep the latest one."""
    data = data.sort(BATCH_TIMESTAMP_COLUMN_NAME)
    data = data.drop(BATCH_TIMESTAMP_COLUMN_NAME)

    return data.unique(subset=primary_key, keep="last")



# TODO: This doesn't run, because it can't find the tools module.
from pathlib import Path
from typing import Any

import requests
from datamodel_code_generator import DataModelType, InputFileType, LiteralType, generate

from seedcase_sprout.internals import _write_json


def get_schema(path: Path) -> Any:
    """Downloads, preprocesses and extracts dataclasses from the Data Package schema."""
    url = "https://datapackage.org/profiles/2.0/datapackage.json"
    response = requests.get(url, timeout=5)
    response.raise_for_status()
    schema = response.json()

    # edit the schema to simplify the generated class structure
    resource_items = schema["properties"]["resources"]["items"]
    del resource_items["oneOf"][0]
    table_schema_properties = resource_items["properties"]["schema"]["properties"]
    del table_schema_properties["fields"]
    table_schema_fkey_items = table_schema_properties["foreignKeys"]["items"]
    table_schema_fkey_items["properties"] = table_schema_fkey_items["oneOf"][0][
        "properties"
    ]
    del table_schema_fkey_items["oneOf"]

    _write_json(schema, path)
    return schema


def generate_dataclasses(path: Path) -> None:
    """Generates dataclasses from the Data Package schema."""
    output_path = Path("tools/generated_properties.py")

    schema = Path(path).read_text()
    return generate(
        schema,
        input_filename=str(path),
        input_file_type=InputFileType.JsonSchema,
        output_model_type=DataModelType.DataclassesDataclass,
        output=output_path,
        custom_template_dir=Path("tools"),
        custom_formatters=["tools.custom_formatters"],
        custom_file_header="# ruff: noqa\n# NOTE: This content was autogenerated, no need to edit it.",  # noqa
        enum_field_as_literal=LiteralType.All,
        use_one_literal_as_default=True,
        use_standard_collections=True,
        use_title_as_name=True,
        use_union_operator=True,
        snake_case_field=True,
        special_field_name_prefix="ignore",
        use_field_description=True,
        use_schema_description=True,
        use_double_quotes=True,
        reuse_model=True,
        force_optional_for_required_fields=True,
    )


if __name__ == "__main__":
    path = Path("tools/schema.json")
    schema = get_schema(path)
    generate_dataclasses(path)
    path.unlink(missing_ok=True)



from pytest import mark

from seedcase_sprout.check_datapackage.check_properties import check_properties

# Without recommendations


def test_passes_matching_properties_with_resources():
    """Should pass properties matching the schema."""
    properties = {
        "name": "a name with spaces",
        "title": "A Title",
        "created": "2024-05-14T05:00:01+00:00",
        "version": "a version",
        "contributors": [{"email": "jane@doe.com"}],
        "sources": [{"email": "jane@doe.com"}],
        "resources": [{"name": "a name", "path": "data.csv"}],
    }

    assert check_properties(properties, check_recommendations=False) == []


def test_fails_properties_without_resources():
    """Should fail properties without resources."""
    properties = {"name": "a name with spaces"}

    errors = check_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "required"
    assert errors[0].json_path == "$.resources"


@mark.parametrize(
    "resources, json_path, num_errors",
    [
        ([], "$.resources", 1),
        ([{}], "$.resources[0].data", 3),
        ([{"name": "a name", "path": "/a/bad/path"}], "$.resources[0].path", 2),
    ],
)
def test_fails_properties_with_bad_resources(resources, json_path, num_errors):
    """Should fail properties with malformed resources."""
    properties = {
        "name": "a name with spaces",
        "resources": resources,
    }

    errors = check_properties(properties, check_recommendations=False)

    assert len(errors) == num_errors
    assert errors[0].json_path == json_path


def test_fails_properties_with_missing_required_fields():
    """Should fail properties with missing required fields."""
    properties = {
        "name": "a name",
        "resources": [{"name": "a name", "path": "data.csv"}],
        "licenses": [{"title": "my license"}],
    }

    errors = check_properties(properties, check_recommendations=False)

    assert len(errors) == 2
    assert all(error.validator == "required" for error in errors)
    assert {error.json_path for error in errors} == {
        "$.licenses[0].name",
        "$.licenses[0].path",
    }


def test_fails_properties_with_bad_type():
    """Should fail properties with a field of the wrong type."""
    properties = {
        "name": 123,
        "resources": [{"name": "a name", "path": "data.csv"}],
    }
    errors = check_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "type"
    assert errors[0].json_path == "$.name"


def test_fails_properties_with_bad_format():
    """Should fail properties with a field of the wrong format."""
    properties = {
        "name": "a name",
        "resources": [{"name": "a name", "path": "data.csv"}],
        "homepage": "not a URL",
    }

    errors = check_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "format"
    assert errors[0].json_path == "$.homepage"


def test_fails_properties_with_pattern_mismatch():
    """Should fail properties with a field that does not match the pattern."""
    properties = {
        "name": "a name",
        "resources": [{"name": "a name", "path": "data.csv"}],
        "contributors": [{"path": "/a/bad/path"}],
    }

    errors = check_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "pattern"
    assert errors[0].json_path == "$.contributors[0].path"


# With recommendations


def test_passes_matching_properties_with_recommendations():
    """Should pass properties matching recommendations."""
    properties = {
        "name": "a-name-with-no-spaces",
        "title": "A Title",
        "id": "123",
        "created": "2024-05-14T05:00:01+00:00",
        "version": "3.2.1",
        "contributors": [{"title": "a contributor"}],
        "sources": [{"title": "a source"}],
        "licenses": [{"name": "a-license"}],
        "resources": [{"name": "a-name-with-no-spaces", "path": "data.csv"}],
    }

    assert check_properties(properties, check_recommendations=True) == []


def test_fails_properties_with_missing_required_fields_with_recommendations():
    """Should fail properties with missing required fields."""
    properties = {
        "resources": [{"name": "a-name-with-no-spaces", "path": "data.csv"}],
    }

    errors = check_properties(properties, check_recommendations=True)

    assert len(errors) == 3
    assert all(error.validator == "required" for error in errors)


def test_fails_properties_violating_recommendations():
    """Should fail properties that do not meet the recommendations."""
    properties = {
        "name": "a name with spaces",
        "id": "123",
        "version": "not semver",
        "contributors": [{"email": "jane@doe.com"}],
        "sources": [{"email": "jane@doe.com"}],
        "licenses": [{"name": "a-license"}],
        "resources": [{"name": "a name with spaces", "path": "data.csv"}],
    }

    errors = check_properties(properties, check_recommendations=True)

    assert len(errors) == 5
    assert {error.json_path for error in errors} == {
        "$.name",
        "$.version",
        "$.contributors[0].title",
        "$.sources[0].title",
        "$.resources[0].name",
    }



from pytest import mark

from seedcase_sprout.check_datapackage import RequiredFieldType
from seedcase_sprout.sprout_checks.check_fields_not_blank import (
    check_fields_not_blank,
)
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)

FIELDS = {"name": RequiredFieldType.str, "tags": RequiredFieldType.list}


@mark.parametrize("index", [None, 2])
def test_no_error_found_in_properties_with_populated_fields(index):
    """Should pass properties with fields populated."""
    properties = {"name": "My name", "tags": ["a", "b"]}

    assert check_fields_not_blank(properties, FIELDS, index) == []


@mark.parametrize("index", [None, 2])
def test_no_error_found_in_properties_with_fields_missing(index):
    """Should pass properties without the specified fields."""
    assert check_fields_not_blank({}, FIELDS, index) == []


@mark.parametrize("index", [None, 2])
def test_error_found_if_properties_have_a_blank_field(index):
    """Should find an error if properties contain a blank field."""
    properties = {"name": "My name", "tags": []}

    errors = check_fields_not_blank(properties, FIELDS, index)

    assert len(errors) == 1
    assert "blank" in errors[0].message
    assert errors[0].json_path == get_json_path_to_resource_field("tags", index)
    assert errors[0].validator == "blank"


@mark.parametrize("index", [None, 2])
def test_error_found_if_properties_have_multiple_blank_fields(index):
    """Should find an error if properties contain multiple blank fields."""
    properties = {"name": "", "tags": []}

    errors = check_fields_not_blank(properties, FIELDS, index)

    assert len(errors) == 2
    assert all(error.validator == "blank" for error in errors)
    assert any(
        error.json_path == get_json_path_to_resource_field("name", index)
        for error in errors
    )
    assert any(
        error.json_path == get_json_path_to_resource_field("tags", index)
        for error in errors
    )



from typing import Any

from seedcase_sprout.check_datapackage import CheckError, RequiredFieldType
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)

CHECKS_REQUIRED_ERROR_MESSAGE = "'{field_name}' is a required property"


def check_fields_present(
    properties: dict[str, Any],
    required_fields: dict[str, RequiredFieldType],
    index: int | None = None,
) -> list[CheckError]:
    """Checks that all fields in `required_fields` are present.

    For resource properties, an index may be supplied, if the resource properties are
    part of a set of package properties.

    Args:
        properties: The properties to check.
        required_fields: The set of required fields and their types.
        index: The index of the resource properties. Defaults to None.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    return [
        CheckError(
            message=CHECKS_REQUIRED_ERROR_MESSAGE.format(field_name=field),
            json_path=get_json_path_to_resource_field(field, index),
            validator="required",
        )
        for field in required_fields.keys()
        if properties.get(field) is None
    ]



from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from unittest.mock import patch
from uuid import UUID
from zoneinfo import ZoneInfo

import time_machine
from pytest import mark

from seedcase_sprout.properties import (
    ConstraintsProperties,
    ContributorProperties,
    FieldProperties,
    LicenseProperties,
    PackageProperties,
    ReferenceProperties,
    ResourceProperties,
    SourceProperties,
    TableSchemaForeignKeyProperties,
    TableSchemaProperties,
)


@mark.parametrize(
    "cls",
    [
        ContributorProperties,
        LicenseProperties,
        SourceProperties,
        ReferenceProperties,
        TableSchemaForeignKeyProperties,
        ConstraintsProperties,
        FieldProperties,
        TableSchemaProperties,
        ResourceProperties,
        PackageProperties,
    ],
)
def test_initiated_class_only_contains_none_values(cls):
    """When a Properties class is instantiated, all values should be None."""
    assert all(value is None for value in asdict(cls()).values())


def test_compact_dict_generates_empty_dictionary_when_no_args_given():
    """Should return an empty dictionary, when no arguments are given to Properties
    class."""
    # Since this is a test of the base class, it's enough to test only one subclass
    assert ResourceProperties().compact_dict == {}


def test_compact_dict_preserves_only_non_none_values():
    """Should return a dictionary with only non-None values when args given."""
    # Since this is a test of the base class, it's enough to test only one subclass

    # Given
    properties = PackageProperties(name="package-1", version="3.2.1")

    # When, then
    assert properties.compact_dict == {"name": "package-1", "version": "3.2.1"}


def test_compact_dict_removes_none_values_in_nested_objects():
    """Given properties with a nested structure, should return a dictionary with only
    non-None values"""
    properties = PackageProperties(
        resources=[
            ResourceProperties(
                schema=TableSchemaProperties(
                    fields=[
                        FieldProperties(
                            constraints=ConstraintsProperties(
                                json_schema={"test": "test"}
                            )
                        )
                    ]
                ),
            ),
        ],
    )

    assert properties.compact_dict == {
        "resources": [
            {"schema": {"fields": [{"constraints": {"jsonSchema": {"test": "test"}}}]}}
        ]
    }


@patch("seedcase_sprout.properties.uuid4", return_value=UUID(int=1))
@time_machine.travel(datetime(2024, 5, 14, 5, 0, 1, tzinfo=ZoneInfo("UTC")), tick=False)
def test_creates_package_properties_with_correct_defaults(mock_uuid):
    """`from_default` should create `PackageProperties` with correct defaults for id,
    version, and created."""
    properties = PackageProperties.from_default()

    assert properties.id == str(mock_uuid())
    assert properties.version == "0.1.0"
    assert properties.created == "2024-05-14T05:00:01+00:00"


@time_machine.travel(datetime(2024, 5, 14, 5, 0, 1, tzinfo=ZoneInfo("UTC")), tick=False)
def test_allows_overriding_defaults():
    """It should be possible to override defaults with non-None values when creating
    `PackageProperties` using `from_default`."""
    properties = PackageProperties.from_default(id="abc", name="my-name", created=None)

    assert properties.name == "my-name"
    assert properties.id == "abc"
    assert properties.version == "0.1.0"
    assert properties.created == "2024-05-14T05:00:01+00:00"


@mark.parametrize(
    "dict, expected_properties",
    [
        ({"familyName": "Doe"}, ContributorProperties(family_name="Doe")),
        ({"name": "a licence"}, LicenseProperties(name="a licence")),
        ({"title": "a source"}, SourceProperties(title="a source")),
        ({"resource": "a resource"}, ReferenceProperties(resource="a resource")),
        ({"fields": ["a field"]}, TableSchemaForeignKeyProperties(fields=["a field"])),
        ({"required": False}, ConstraintsProperties(required=False)),
        ({"name": "a field name"}, FieldProperties(name="a field name")),
        ({"fieldsMatch": ["exact"]}, TableSchemaProperties(fields_match=["exact"])),
        (
            {
                "name": "resource-name",
                "licenses": [{"name": "MIT"}],
            },
            ResourceProperties(
                name="resource-name",
                licenses=[LicenseProperties(name="MIT")],
            ),
        ),
        (
            {"version": "1.0.0", "contributors": [{"familyName": "Doe"}]},
            PackageProperties(
                version="1.0.0", contributors=[ContributorProperties(family_name="Doe")]
            ),
        ),
    ],
)
def test_transforms_dict_to_properties(dict, expected_properties):
    """Should transform a (nested) dictionary to a properties object."""
    properties_cls = type(expected_properties)
    properties = properties_cls.from_dict(dict)

    assert properties == expected_properties


@mark.parametrize(
    "resource_properties, path",
    [
        (ResourceProperties(name="space in name"), None),
        (
            ResourceProperties(name="test-resource"),
            str(Path("resources", "test-resource", "data.parquet")),
        ),
        (
            ResourceProperties(name="test-resource", path="some/path"),
            str(Path("resources", "test-resource", "data.parquet")),
        ),
    ],
)
def test_autogenerates_resource_data_path(resource_properties, path):
    """Should autogenerate the resource path from the resource name after object
    creation (if the name is correct)."""
    assert resource_properties.path == path


def test_resource_path_included_in_compact_dict():
    """`compact_dict` should include the autogenerated resource path."""
    resource_properties = ResourceProperties(name="test-resource")

    assert resource_properties.compact_dict == {
        "name": "test-resource",
        "path": str(Path("resources", "test-resource", "data.parquet")),
    }



import polars as pl
from polars.testing import assert_frame_equal
from pytest import fixture, raises

from seedcase_sprout.constants import BATCH_TIMESTAMP_COLUMN_NAME
from seedcase_sprout.examples import example_resource_properties
from seedcase_sprout.join_resource_batches import join_resource_batches
from seedcase_sprout.properties import (
    ResourceProperties,
)


@fixture
def data_list() -> list[pl.DataFrame]:
    return [
        pl.DataFrame(
            {
                "id": [0, 1],
                "name": ["anne", "belinda"],
                "value": [0.0, 1.1],
                # timestamp col from `read_resource_batches`
                BATCH_TIMESTAMP_COLUMN_NAME: ["2025-03-26T100000Z"] * 2,
            }
        ),
        pl.DataFrame(
            {
                "id": [2, 3, 0, 0, 0],
                "name": ["catherine", "dorothy", "anne", "anne", "alberta"],
                "value": [2.2, 3.3, 0.0, 9.9, 0.0],
                # timestamp col from `read_resource_batches` (different year than above)
                BATCH_TIMESTAMP_COLUMN_NAME: ["2024-03-26T100000Z"] * 5,
            }
        ),
    ]


@fixture
def resource_properties() -> ResourceProperties:
    """Fixture for resource_properties."""
    return example_resource_properties()


def test_batches_are_joined_correctly(data_list, resource_properties):
    """Test that the batches are joined correctly."""
    # Given
    resource_properties.schema.primary_key = "id"

    # When
    joined_batches = join_resource_batches(
        data_list=data_list,
        resource_properties=resource_properties,
    )

    # Then
    expected_joined_batches = pl.DataFrame(
        {
            # only one row with id 0 is kept, the latest one (it is unique in terms of
            # id, ignoring timestamp)
            "id": [1, 2, 3, 0],
            "name": ["belinda", "catherine", "dorothy", "anne"],
            "value": [1.1, 2.2, 3.3, 0.0],
        }
    )

    assert_frame_equal(
        joined_batches,
        expected_joined_batches,
        check_row_order=False,
    )


def test_batches_are_joined_correct_with_no_primary_key(data_list, resource_properties):
    """Duplicate rows (all cells are identical except timestamp) are removed when there
    isn't a primary key."""
    # Given
    resource_properties.schema.primary_key = None

    # When
    joined_batches = join_resource_batches(
        data_list=data_list,
        resource_properties=resource_properties,
    )
    # Then
    expected_joined_batches = pl.DataFrame(
        {
            # three rows with id 0 are kept (they are unique, ignoring timestamp)
            "id": [1, 2, 3, 0, 0, 0],
            "name": [
                "belinda",
                "catherine",
                "dorothy",
                "anne",
                "anne",
                "alberta",
            ],
            "value": [1.1, 2.2, 3.3, 0.0, 9.9, 0.0],
        }
    )

    assert_frame_equal(
        joined_batches,
        expected_joined_batches,
        check_row_order=False,
    )


def test_batches_joined_correctly_when_primary_key_is_multiple_fields(
    data_list, resource_properties
):
    """Batches are joined correctly when the primary key is multiple fields."""
    # Given
    resource_properties.schema.primary_key = ["id", "value"]

    # When
    joined_batches = join_resource_batches(
        data_list=data_list,
        resource_properties=resource_properties,
    )
    # Then
    expected_joined_batches = pl.DataFrame(
        {
            # two rows with id 0 are kept (they are unique in terms of id and value,
            # ignoring timestamp)
            "id": [1, 2, 3, 0, 0],
            "name": [
                "belinda",
                "catherine",
                "dorothy",
                "anne",
                "anne",
            ],
            "value": [1.1, 2.2, 3.3, 0.0, 9.9],
        }
    )

    assert_frame_equal(joined_batches, expected_joined_batches, check_row_order=False)


def test_throws_error_with_data_of_different_shapes(data_list, resource_properties):
    """Test that an error is raised when the data have different shapes."""
    # Given, when
    data_list.append(
        pl.DataFrame(
            {
                "id": [2],
                "name": ["bertha"],
                # value column is missing
                BATCH_TIMESTAMP_COLUMN_NAME: ["2024-03-26T100000Z"],
            }
        )
    )

    # Then
    with raises(pl.exceptions.ShapeError):
        join_resource_batches(
            data_list=data_list,
            resource_properties=resource_properties,
        )


def test_throws_error_with_non_matching_data_types(data_list, resource_properties):
    """Test that an error is raised when the data types don't match."""
    # Given, when
    data_list.append(
        pl.DataFrame(
            {
                "id": [2],
                "name": ["bertha"],
                "value": [1.1],
                BATCH_TIMESTAMP_COLUMN_NAME: ["2024-03-26T100000Z"],
            },
            schema={
                "id": pl.Int64,
                "name": pl.String,
                "value": pl.Object,  # different type than the other dataframes
                BATCH_TIMESTAMP_COLUMN_NAME: pl.String,
            },
        ),
    )

    # Then
    with raises(pl.exceptions.SchemaError):
        join_resource_batches(
            data_list=data_list,
            resource_properties=resource_properties,
        )


def test_throws_error_with_non_matching_column_names(data_list, resource_properties):
    """Test that an error is raised when the column names don't match."""
    # Given, when
    data_list.append(
        pl.DataFrame(
            {
                "id": [2],
                "unexpected_column_name": ["bertha"],
                "value": [1.1],
                BATCH_TIMESTAMP_COLUMN_NAME: ["2024-03-26T100000Z"],
            },
        )
    )

    # Then
    with raises(pl.exceptions.ShapeError):
        join_resource_batches(
            data_list=data_list,
            resource_properties=resource_properties,
        )


def test_single_dataframe_in_data_list(data_list, resource_properties):
    """Test that a single DataFrame is returned as is (except row order)."""
    # Given
    data_list = [data_list[0]]

    # When
    joined_batches = join_resource_batches(
        data_list=data_list,
        resource_properties=resource_properties,
    )

    # Then
    expected_joined_batches = data_list[0].drop(BATCH_TIMESTAMP_COLUMN_NAME)
    assert_frame_equal(joined_batches, expected_joined_batches, check_row_order=False)


def test_throws_error_with_empty_data_list(resource_properties):
    """Should throw an informative error if an empty data list is provided."""
    with raises(ValueError) as error:
        join_resource_batches([], resource_properties)

    assert resource_properties.name in str(error)



from pathlib import Path

from seedcase_sprout.internals.to import _to_snake_case


def _create_resource_data_path(resource_name: str) -> str:
    """Creates a stringified relative path to the resource data file based on the name.

    Args:
        resource_name: The name of the resource.

    Returns:
        The relative path from the package root to the resource data file.
            E.g., "resources/test-resource/data.parquet"
    """
    return str(Path("resources", resource_name, "data.parquet"))


def _create_resource_properties_script_filename(resource_name: str = "") -> str:
    """Creates the filename for the resource properties script from the resource name.

    Args:
        resource_name: The name of the resource. Defaults to "".

    Returns:
        The filename.
    """
    return f"resource_properties{resource_name and '_'}{_to_snake_case(resource_name)}"



import re
from typing import Any

from seedcase_sprout.check_datapackage.constants import NAME_PATTERN


def _is_resource_name_correct(resource_name: Any) -> bool:
    """Checks if the given resource name is correct.

    Args:
        resource_name: The resource name to check.

    Returns:
        Whether the resource name is correct.
    """
    return isinstance(resource_name, str) and bool(
        re.match(NAME_PATTERN, resource_name)
    )



from seedcase_sprout.check_datapackage import CheckError


def exclude_non_sprout_resource_errors(
    errors: list[CheckError],
) -> list[CheckError]:
    """Filters out resource errors that are not relevant for Sprout.

    Errors filtered out:
      - inline `data` required but missing
      - `path` is not of type array

    Args:
        errors: The full error list.

    Returns:
        The filtered error list.
    """
    return [
        error
        for error in errors
        if not (error.validator == "required" and error.json_path.endswith(".data"))
        and not (
            error.validator == "type"
            and error.json_path.endswith(".path")
            and error.message.endswith("not of type 'array'")
        )
    ]



from json import loads
from pathlib import Path
from typing import Any


def _read_json(path: Path) -> dict[str, Any]:
    """Reads the contents of a JSON file into an object.

    Args:
        path: The path to the file to load.

    Returns:
        The contents of the file as an object.

    Raises:
        JSONDecodeError: If the contents of the file cannot be de-serialised as JSON.
        TypeError: If the object in the file is not a dictionary.
    """
    loaded_object = loads(path.read_text())
    if not isinstance(loaded_object, dict):
        raise TypeError(
            f"Expected {path} to contain a JSON dictionary object "
            f"but found {type(loaded_object)}."
        )
    return loaded_object



from pytest import mark, raises

from seedcase_sprout.create_resource_properties_script import (
    create_resource_properties_script,
)
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import (
    FieldProperties,
    ResourceProperties,
    TableSchemaProperties,
)
from tests.load_properties import load_properties


@mark.parametrize("args", [[], [None] * 3])
def test_creates_script_with_empty_resource_properties(tmp_cwd, args):
    """Should create a script with an empty `ResourceProperties` object."""
    script_path = create_resource_properties_script(*args)

    assert script_path == PackagePath().resource_properties_script()
    properties = load_properties(script_path, "resource_properties")
    assert properties == ResourceProperties(
        name="",
        title="",
        description="",
        type="table",
        format="parquet",
        mediatype="application/parquet",
        schema=TableSchemaProperties(fields=[FieldProperties(name="", type="")]),  # type: ignore[arg-type]
    )


@mark.parametrize("resource_name", ["my_resource", "my.resource", "my-resource"])
def test_creates_script_with_name_and_fields(tmp_cwd, resource_name):
    """Should create a script with resource name and fields."""
    fields = [
        FieldProperties(name="field1", type="string"),
        FieldProperties(name="field2", type="datetime"),
    ]

    script_path = create_resource_properties_script(
        resource_name=resource_name,
        fields=fields,
    )

    properties = load_properties(script_path, "resource_properties_my_resource")
    assert properties == ResourceProperties(
        name=resource_name,
        title="",
        description="",
        type="table",
        format="parquet",
        mediatype="application/parquet",
        schema=TableSchemaProperties(fields=fields),
    )


def test_works_with_custom_path(tmp_path):
    """Should work with a custom path."""
    script_path = create_resource_properties_script(path=tmp_path)

    assert script_path == PackagePath(tmp_path).resource_properties_script()


def test_incorrect_resource_name_raises_error(tmp_cwd):
    """Should raise a error if an incorrect resource name is provided."""
    with raises(ValueError, match="resource name"):
        create_resource_properties_script(resource_name="spaces in name")


@mark.parametrize("args", [[], ["my-resource"]])
def test_existing_script_not_overwritten(tmp_cwd, args):
    """If a script already exists for a resource, it should not be overwritten."""
    script_path = create_resource_properties_script(*args)
    script_path.write_text("test")

    create_resource_properties_script(*args)

    assert script_path.read_text() == "test"



import os

from pytest import fixture


@fixture
def tmp_cwd(tmp_path):
    """Temporarily changes the working directory to a temporary path.

    This fixture changes the current working directory to a temporary one
    (provided by pytest's `tmp_path`) for the duration of the test. After the
    test finishes, the original working directory is restored.

    Yields:
        The temporary path that becomes the current working directory.

    Examples:
        ```{python}
        def test_something(tmp_cwd):
            # cwd is now a temporary directory
            ...
        ```
    """
    original = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(original)



from pytest import mark, raises

from seedcase_sprout.check_datapackage.check_error import CheckError

check_error = CheckError(
    message="There was an error!", json_path="a.b.field", validator="a-validator"
)


def test_error_stringified_correctly():
    """Should stringify error correctly."""
    assert (
        str(check_error)
        == "Error at `a.b.field` caused by `a-validator`: There was an error!"
    )


def test_error_represented_correctly():
    """Should generate the developer-friendly representation correctly."""
    assert repr(check_error) == (
        "CheckError(message='There was an error!', json_path='a.b.field', "
        "validator='a-validator')"
    )


@mark.parametrize(
    "error, other, expected",
    [
        (CheckError("msg1", "path1", "v1"), CheckError("msg1", "path1", "v1"), True),
        (CheckError("msg1", "path1", "v1"), CheckError("msg2", "path1", "v1"), False),
        (CheckError("msg1", "path1", "v1"), CheckError("msg1", "path2", "v1"), False),
        (CheckError("msg1", "path1", "v1"), CheckError("msg1", "path1", "v2"), False),
        (CheckError("msg1", "path1", "v1"), None, False),
        (CheckError("msg1", "path1", "v1"), "not a CheckError", False),
    ],
)
def test_equality_checked_correctly(error, other, expected):
    """Should compare errors for equality correctly."""
    assert (error == other) is expected
    assert (other == error) is expected


@mark.parametrize(
    "error, other, expected",
    [
        (CheckError("msg A", "path1", "v1"), CheckError("msg A", "path2", "v1"), True),
        (CheckError("msg A", "path1", "v1"), CheckError("msg A", "path1", "v2"), True),
        (CheckError("msg A", "path1", "v1"), CheckError("msg B", "path1", "v1"), True),
        (CheckError("msg A", "path2", "v1"), CheckError("msg A", "path1", "v1"), False),
        (CheckError("msg A", "path1", "v2"), CheckError("msg A", "path1", "v1"), False),
        (CheckError("msg B", "path1", "v1"), CheckError("msg A", "path1", "v1"), False),
    ],
)
def test_lt_checked_correctly_for_different_errors(error, other, expected):
    """Should find the smaller (the one that comes first) of two different errors."""
    assert (error < other) is expected
    assert (other < error) is not expected


def test_lt_checked_correctly_for_same_error():
    """Should not consider an error smaller than itself."""
    assert not (check_error < check_error)


@mark.parametrize(
    "error, other",
    [
        (CheckError("msg", "path", "val"), None),
        (CheckError("msg", "path", "val"), "not a CheckError"),
    ],
)
def test_lt_raises_error_for_wrong_type(error, other):
    """Should raise an error when comparing an error with an object of a different
    type."""
    with raises(TypeError):
        error < other



from typing import Any

from seedcase_sprout.check_datapackage import CheckError
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)

CHECKS_TYPE_ERROR_MESSAGE = "{field_value} is not of type '{field_type}'"


def check_resource_path_string(
    properties: dict[str, Any], index: int | None = None
) -> list[CheckError]:
    """Checks that the `path` field of a set of resource properties is of type string.

    Args:
        properties: The resource properties.
        index: The index of the resource properties. Defaults to None.

    Returns:
        A list of errors. An empty list if no error was found.
    """
    path = properties.get("path", "")
    if isinstance(path, str):
        return []

    return [
        CheckError(
            message=CHECKS_TYPE_ERROR_MESSAGE.format(
                field_value=path, field_type="string"
            ),
            json_path=get_json_path_to_resource_field("path", index),
            validator="type",
        )
    ]



from pathlib import Path

from seedcase_sprout.check_properties import check_properties
from seedcase_sprout.internals import _check_is_file, _read_json
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import PackageProperties


def read_properties(path: Path | None = None) -> PackageProperties:
    """Read in the properties from the `datapackage.json` file.

    Reads the `datapackage.json` file, checks that it is correct, and then
    outputs a `PackageProperties` object.

    Args:
        path: The path to the `datapackage.json` file. Use `PackagePath().properties()`
            to help get the correct path. If no path is provided, this function looks
            for the `datapackage.json` file in the current working directory.

    Returns:
        A `PackageProperties` object with the properties from the
            `datapackage.json` file.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            sp.read_properties()
        ```

    Raises:
        FileNotFound: If the `datapackage.json` file doesn't exist.
        JSONDecodeError: If the `datapackage.json` file couldn't be read.
    """
    path = path or PackagePath().properties()
    _check_is_file(path)
    properties_dict = _read_json(path)
    properties = PackageProperties.from_dict(properties_dict)
    check_properties(properties)
    return properties



import re


def _to_snake_case(name: str) -> str:
    """Creates a snake-case version of a package or resource name.

    Args:
        name: The package or resource name.

    Returns:
        The package or resource name in snake case.

    """
    return name.replace(".", "_").replace("-", "_").lower()


def _to_camel_case(text: str) -> str:
    """Converts snake case to camel case.

    Args:
        text: The snake-case string to convert.

    Returns:
        The converted string in camel case.
    """
    first_part, *remaining_parts = text.split("_")
    return first_part + "".join(part.title() for part in remaining_parts)


def _to_dedented(text: str | None) -> str | None:
    """Dedents text by removing leading whitespace and tabs from each line.

    If it is not indented, it will be returned as is.
    If the first line starts with newlines, those will be removed.

    Args:
        text: The text to dedent.

    Returns:
        The dedented text or `None` if the input is `None`.
    """
    if text is None:
        return None

    # Remove leading whitespace or tabs from each line
    dedented = re.sub(r"^[ \t]+", "", text, flags=re.MULTILINE)
    # Remove leading newlines from the very start
    return dedented.lstrip("\n")



from pytest import mark

from seedcase_sprout.check_datapackage import RequiredFieldType
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)


@mark.parametrize(
    "type,value",
    [
        (RequiredFieldType.str, ""),
        (RequiredFieldType.list, []),
        ("int", None),
        (None, None),
        ("something else", None),
    ],
)
def test_returns_expected_blank_value_for_each_type(type, value):
    """Should return the expected blank value for each type."""
    assert get_blank_value_for_type(type) == value



"""This module includes the dataclasses for the properties of a data package.

The properties are based on the Frictionless Data Package specification. They are used
as input for creating and editing the properties of data packages and data resources and
are intended to help users with the correct structure and content of the properties.
"""
# NOTE: This content is modified from the auto-generated
# `generate_properties/generated_properties.py` file. Update the auto-generated
# properties file to add more dataclasses and move them into this file.

from abc import ABC
from dataclasses import asdict, dataclass
from typing import Any, Literal, Self
from uuid import uuid4

from dacite import Config, from_dict

from seedcase_sprout.internals import (
    _create_resource_data_path,
    _get_iso_timestamp,
    _to_camel_case,
)
from seedcase_sprout.sprout_checks.is_resource_name_correct import (
    _is_resource_name_correct,
)


@dataclass
class Properties(ABC):
    """An abstract base class for all `*Properties` classes holding common logic."""

    @property
    def compact_dict(self) -> dict[str, Any]:
        """Converts the dataclass `*Properties` object to a `camelCase` dictionary.

        Applies recursively to nested `*Properties` objects. Also removes any keys with
        None values.

        Returns:
            A dictionary representation of the `*Properties` object with only non-None
                values.
        """
        return asdict(
            obj=self,
            dict_factory=lambda tuples: {
                _to_camel_case(key): value for key, value in tuples if value is not None
            },
        )

    @classmethod
    def from_dict(cls: type[Self], data: dict[str, Any]) -> Self:
        """Creates a dataclass `*Properties` object filled with data from a dictionary.

        Args:
            cls: The class to create the `*Properties` object from.
            data: The data to fill the `*Properties` object with.

        Returns:
            A `*Properties` object with the properties from the dictionary.
        """
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                # Expect keys in the input to be in camel case
                convert_key=_to_camel_case,
            ),
        )


@dataclass
class ContributorProperties(Properties):
    """The people or organizations who contributed to this data package.

    Creates a dataclass object with all the necessary properties for a
    contributor. This would be given in the `contributors` field of a
    `PackageProperties` object.

    Attributes:
        title (str | None): The name of the contributor.
        path (str | None): A fully qualified URL pointing to a relevant
            location online for the contributor.
        email (str | None): An email address.
        given_name (str | None): The name a person has been
            given, if the contributor is a person.
        family_name (str | None): The familial name that a person
            inherits, if the contributor is a person.
        organization (str | None): An organizational affiliation for this
            contributor.
        roles (list[str] | None): An array of strings describing the roles of the
            contributor.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        print(sp.ContributorProperties())
        print(sp.ContributorProperties(title="Amir Smith"))
        ```
    """

    title: str | None = None
    path: str | None = None
    email: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    organization: str | None = None
    roles: list[str] | None = None


@dataclass
class LicenseProperties(Properties):
    """The license(s) under which the package or resource is provided.

    Creates a dataclass object with all the necessary properties for a
    license, so that it can be added to the `licenses` field of a
    `PackageProperties` object.

    Attributes:
        name (str | None): Must be an Open Definition license identifier,
            see http://licenses.opendefinition.org/
        path (str | None): A fully qualified URL, or a POSIX file path.
        title (str | None): A human-readable title.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        print(sp.LicenseProperties())
        ```
    """

    name: str | None = None
    path: str | None = None
    title: str | None = None


@dataclass
class SourceProperties(Properties):
    """The raw sources for this data package.

    Attributes:
        title (str | None): The title of the source (e.g. document or organization).
        path (str | None): A fully qualified URL, or a POSIX file path.
        email (str | None): An email address.
        version (str | None): The version of the source.
    """

    title: str | None = None
    path: str | None = None
    email: str | None = None
    version: str | None = None


@dataclass
class ReferenceProperties(Properties):
    """The destination part of a foreign key.

    Attributes:
        resource (str | None): The name of the resource within the current data
            package where the `fields` are located.
        fields (list[str] | None): An array of strings of the same length as
            `TableSchemaForeignKeyProperties.fields`, specifying the field (or fields)
            that form the destination part of the foreign key.
    """

    resource: str | None = None
    fields: list[str] | None = None


@dataclass
class TableSchemaForeignKeyProperties(Properties):
    """A foreign key in a table schema.

    A foreign key is a reference where values in a field (or fields) on the table
    ("resource" in data package terminology) described by the table schema connect to
    values in a field (or fields) on this or a separate table (resource).

    Attributes:
        fields (list[str] | None): An array of strings specifying the field (or
            fields) on this resource that form the source part of the foreign key.
        reference (ReferenceProperties | None): An object specifying the
            destination part of the foreign key.
    """

    fields: list[str] | None = None
    reference: ReferenceProperties | None = None


# Allowed types for a field in a table schema.
FieldType = Literal[
    "string",
    "number",
    "integer",
    "boolean",
    "object",
    "array",
    "datetime",
    "date",
    "time",
    "year",
    "yearmonth",
    "duration",
    "geopoint",
    "geojson",
    "any",
]


@dataclass
class ConstraintsProperties(Properties):
    """A dataclass that expresses constraints for validating field values.

    A constraint is a rule that dictates the given values, or range of values,
    that a variable or column can have in a dataset. For instance, a constraint
    for an "age" column could be that it must be greater than 0 but less than 120.

    Attributes:
        required (bool | None): Indicates whether a property must have a
            value for each instance.
        unique (bool | None): When `true`, each value for the property
            must be unique.
        pattern (str | None): A regular expression pattern to test each
            value of the property against, where a truthy response indicates
            validity.
        enum (list[Any] | None): The value of the field must exactly match one of
            the values in the `enum` array.
        min_length (int | None): An integer that specifies the minimum
            length of a value.
        max_length (int | None): An integer that specifies the maximum
            length of a value.
        minimum (str | float | int | None): Specifies a minimum value for a field.
        maximum (str | float | int | None): Specifies a maximum value for a field.
        exclusive_minimum (str | float | int | None): Specifies an exclusive minimum
            value for a field.
        exclusive_maximum (str | float | int | None): Specifies an exclusive maximum
            value for a field.
        json_schema (dict[str, Any] | None): A valid JSON schema object to
            validate field values. If a field value conforms to the provided
            JSON schema then this field value is valid.
    """

    required: bool | None = None
    unique: bool | None = None
    pattern: str | None = None
    enum: list[Any] | None = None
    min_length: int | None = None
    max_length: int | None = None
    minimum: str | float | int | None = None
    maximum: str | float | int | None = None
    exclusive_minimum: str | float | int | None = None
    exclusive_maximum: str | float | int | None = None
    json_schema: dict[str, Any] | None = None


@dataclass
class FieldProperties(Properties):
    """A field in a table schema.

    Provides human-readable documentation as well as additional information that can
    be used to validate the field or create a user interface for data entry.

    Attributes:
        name (str | None): A name for this field. Must be unique amongst other field
            names in this table schema.
        title (str | None): A human readable label or title for this field.
        type (FieldType | None): The data type of this field.
        format (str | None): The format for this field.
        description (str | None): A text description for this field.
        example (str | None): An example value for this field.
        constraints (ConstraintsProperties | None): The constraints applicable to
            this field.
        categories (list[str] | list[int] | None): A finite set of possible values
            for this field.
        categories_ordered (bool | None): Specifies whether the order of appearance
            of the values in the `categories` property should be regarded as their
            natural order.
    """

    name: str | None = None
    title: str | None = None
    type: FieldType | None = None
    format: str | None = None
    description: str | None = None
    example: str | None = None
    constraints: ConstraintsProperties | None = None
    categories: list[str] | list[int] | None = None
    categories_ordered: bool | None = None


# Allowed strategies for matching fields in the table schema to fields the data source.
FieldsMatchType = Literal["exact", "equal", "subset", "superset", "partial"]


@dataclass
class TableSchemaProperties(Properties):
    """A table schema for a data resource.

    Table schema is a simple language- and implementation-agnostic way to declare a
    schema for tabular data. Table schema is well suited for use cases around handling
    and validating tabular data in text formats such as CSV, but its utility extends
    well beyond this core usage, towards a range of applications where data benefits
    from a portable schema format.

    Attributes:
        fields (list[FieldProperties] | None): Specifies the fields in this table
            schema.
        fields_match (list[FieldsMatchType] | None): Specifies how fields in the table
            schema match the fields in the data source.
        primary_key (list[str] | str | None): A primary key is a field name or an array
            of field names, whose values must uniquely identify each row in the table.
        unique_keys (list[list[str]] | None): A field or a set of fields that are
            required to have unique logical values in each row in the table.
        foreign_keys (list[TableSchemaForeignKeyProperties] | None): A reference where
            values in a field (or fields) on the table (resource) described by this
            table schema connect to values in a field (or fields) on this or a separate
            table (resource).

    Examples:
        ```{python}
        import seedcase_sprout as sp
        print(sp.TableSchemaProperties(primary_key="id"))
        ```
    """

    fields: list[FieldProperties] | None = None
    fields_match: list[FieldsMatchType] | None = None
    primary_key: list[str] | str | None = None
    unique_keys: list[list[str]] | None = None
    foreign_keys: list[TableSchemaForeignKeyProperties] | None = None


@dataclass
class ResourceProperties(Properties):
    """A data resource.

    A simple format to describe and package a single data resource such as an
    individual table or file. The essence of a data resource is a locator for
    the data it describes. A range of other properties can be declared to provide a
    richer set of metadata.

    Creates a dataclass object with all the necessary properties for a resource,
    which would be given in the `resources` field of a `PackageProperties`
    dataclass.

    Attributes:
        name (str | None): A simple name or identifier to be used for this resource.
            Should consist only of lowercase English alphanumeric characters plus
            characters in `.-_`.
        path (str | None): A path pointing to the data for this resource.
        type (Literal['table'] | None): Specifies the type of the resource.
        title (str | None): A human-readable title.
        description (str | None): A text description. Markdown is encouraged.
        sources (list[SourceProperties] | None): The raw sources for this resource.
        licenses (list[LicenseProperties] | None): The license(s) under which the
            resource is published.
        format (str | None): The file format of this resource. Expected to be the
            standard file extension.
        mediatype (str | None): The media type of this resource. Can be any
            valid media type listed with
            [IANA](https://www.iana.org/assignments/media-types/media-types.xhtml).
        encoding (str | None): The file encoding of this resource.
        bytes (int | None): The size of this resource in bytes.
        hash (str | None): The MD5 hash of this resource. Indicate other
            hashing algorithms with the {algorithm}:{hash} format.
        schema (TableSchemaProperties | None): A table schema for the resource data,
            compliant with the table schema specification.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        print(sp.ResourceProperties())
        print(sp.ResourceProperties(name="blood-samples", title="Blood samples"))
        ```
    """

    name: str | None = None
    path: str | None = None
    type: Literal["table"] | None = None
    title: str | None = None
    description: str | None = None
    sources: list[SourceProperties] | None = None
    licenses: list[LicenseProperties] | None = None
    format: str | None = None
    mediatype: str | None = None
    encoding: str | None = None
    bytes: int | None = None
    hash: str | None = None
    schema: TableSchemaProperties | None = None

    def __post_init__(self) -> None:
        """Generates the path from the resource name after object creation."""
        self.path = (
            _create_resource_data_path(str(self.name))
            if _is_resource_name_correct(self.name)
            else None
        )


@dataclass
class PackageProperties(Properties):
    """Properties for a data package.

    A simple container format for describing a coherent collection of data in a single
    "package". It provides the basis for convenient delivery, installation and
    management of datasets.

    Creates a dataclass object with all the necessary properties for the package.

    Attributes:
        name (str | None): A simple name or identifier to be used for this package.
            Should consist only of lowercase English alphanumeric characters plus
            characters in `.-_`.
        id (str | None): The unique identifier of this package.
        title (str | None): A human-readable title.
        description (str | None): A text description. Markdown is encouraged.
        homepage (str | None): The home on the web that is related to this package.
        version (str | None): A version string identifying the version of this package.
        created (str | None): The datetime on which this package was created.
        contributors (list[ContributorProperties] | None): The people or organizations
            who contributed to this package.
        keywords (list[str] | None): A list of keywords that describe this package.
        image (str | None): An image to represent this package.
        licenses (list[LicenseProperties] | None): The license(s) under which this
            package is published.
        resources (list[ResourceProperties] | None): Specifies the data resources
            in this data package, each compliant with the data resource specification.
        sources (list[SourceProperties] | None): The raw sources for this data
            package.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        print(sp.PackageProperties())
        print(sp.PackageProperties(name="diabetes-cohort", title="Diabetes Cohort"))
        print(sp.PackageProperties(licenses=[sp.LicenseProperties(name="ODC-BY-1.0")]))

        # To allow multiline strings, use dedent.
        from textwrap import dedent
        print(sp.PackageProperties(
            title="Birds of North America",
            description=dedent('''
                # Markdown header

                A dataset of bird sightings. With some **bolding**.
                '''
            )
        ))
        ```
    """

    name: str | None = None
    id: str | None = None
    title: str | None = None
    description: str | None = None
    homepage: str | None = None
    version: str | None = None
    created: str | None = None
    contributors: list[ContributorProperties] | None = None
    keywords: list[str] | None = None
    image: str | None = None
    licenses: list[LicenseProperties] | None = None
    resources: list[ResourceProperties] | None = None
    sources: list[SourceProperties] | None = None

    @classmethod
    def from_default(
        cls,
        *,
        name: str | None = None,
        id: str | None = None,
        title: str | None = None,
        description: str | None = None,
        homepage: str | None = None,
        version: str | None = None,
        created: str | None = None,
        contributors: list[ContributorProperties] | None = None,
        keywords: list[str] | None = None,
        image: str | None = None,
        licenses: list[LicenseProperties] | None = None,
        resources: list[ResourceProperties] | None = None,
        sources: list[SourceProperties] | None = None,
    ) -> Self:
        """Creates a `PackageProperties` dataclass with default values.

        Default values (`id`, `version`, and `created`) can be overridden and unset
        values can be set using keyword arguments.

        Args:
            name (str | None): A simple name or identifier to be used for this package.
                Should consist only of lowercase English alphanumeric characters plus
                characters in `.-_`.
            id (str | None): The unique identifier of this package.
            title (str | None): A human-readable title.
            description (str | None): A text description. Markdown is encouraged.
            homepage (str | None): The home on the web that is related to this package.
            version (str | None): A version string identifying the version of this
                package.
            created (str | None): The datetime on which this package was created.
            contributors (list[ContributorProperties] | None): The people or
                organizations
                who contributed to this package.
            keywords (list[str] | None): A list of keywords that describe this package.
            image (str | None): An image to represent this package.
            licenses (list[LicenseProperties] | None): The license(s) under which this
                package is published.
            resources (list[ResourceProperties] | None): Specifies the data resources
                in this data package, each compliant with the data resource
                specification.
            sources (list[SourceProperties] | None): The raw sources for this data
                package.

        Returns:
            A `PackageProperties` dataclass with default values.

        Examples:
            ```{python}
            import seedcase_sprout as sp

            sp.PackageProperties.from_default(name="my-package", title="My Package...")
            ```
        """
        return cls(
            name=name,
            id=id or str(uuid4()),
            title=title,
            description=description,
            homepage=homepage,
            version=version or "0.1.0",
            created=created or _get_iso_timestamp(),
            contributors=contributors,
            keywords=keywords,
            image=image,
            licenses=licenses,
            resources=resources,
            sources=sources,
        )



from typing import Any

import seedcase_sprout.check_datapackage as cdp
from seedcase_sprout.internals import (
    _create_resource_data_path,
)
from seedcase_sprout.sprout_checks.check_fields_not_blank import (
    check_fields_not_blank,
)
from seedcase_sprout.sprout_checks.check_fields_present import (
    check_fields_present,
)
from seedcase_sprout.sprout_checks.check_no_inline_data import check_no_inline_data
from seedcase_sprout.sprout_checks.check_resource_path_string import (
    check_resource_path_string,
)
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)
from seedcase_sprout.sprout_checks.is_resource_name_correct import (
    _is_resource_name_correct,
)
from seedcase_sprout.sprout_checks.required_fields import (
    RESOURCE_SPROUT_REQUIRED_FIELDS,
)


def get_sprout_resource_errors(
    properties: dict[str, Any], index: int | None = None
) -> list[cdp.CheckError]:
    """Checks the resource `properties` against Sprout-specific requirements only.

    Args:
        properties: The resource properties.
        index: The index of the resource properties. Defaults to None.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    errors = check_resource_path_string(properties, index)
    errors += _check_resource_path_format(properties, index)
    errors += check_no_inline_data(properties, index)
    errors += check_fields_not_blank(properties, RESOURCE_SPROUT_REQUIRED_FIELDS, index)
    errors += check_fields_present(properties, RESOURCE_SPROUT_REQUIRED_FIELDS, index)
    return errors


def _check_resource_path_format(
    properties: dict[str, Any], index: int | None = None
) -> list[cdp.CheckError]:
    """Checks if the data path in the resource properties has the correct format.

    As the path is constructed from the resource name, its format can only be checked
    if the resource name is correct. Type, required, and blank errors are not flagged
    here to avoid flagging them twice.

    Args:
        properties: The resource properties to check.
        index: The index of the resource properties. Defaults to None.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    name = properties.get("name")
    path = properties.get("path")

    if (
        # Do not check path if name is incorrect
        not _is_resource_name_correct(name)
        # Do not flag type and required errors twice
        or not isinstance(path, str)
        # Do not flag blank errors twice
        or path == ""
    ):
        return []

    expected_path = _create_resource_data_path(str(name))
    if path == expected_path:
        return []

    return [
        cdp.CheckError(
            message=f"Expected the path to be '{expected_path}' but found '{path}'.",
            json_path=get_json_path_to_resource_field("path", index),
            validator="pattern",
        )
    ]



from pathlib import Path
from uuid import uuid4

import polars as pl
from pytest import fixture, mark, raises

from seedcase_sprout.constants import BATCH_TIMESTAMP_COLUMN_NAME
from seedcase_sprout.examples import example_resource_properties
from seedcase_sprout.properties import (
    FieldProperties,
    ResourceProperties,
    TableSchemaProperties,
)
from seedcase_sprout.read_resource_batches import (
    read_resource_batches,
)
from tests.assert_raises_errors import (
    assert_raises_check_errors,
)
from tests.directory_structure_setup import (
    create_test_data_package,
)

batch_data_1 = pl.DataFrame(
    {
        "id": [0, 1, 2],
        "name": ["anne", "belinda", "catherine"],
    }
)

batch_data_2 = pl.DataFrame(
    {
        "id": [3, 4, 5],
        "name": ["dorothy", "figaro", "gabrielle"],
    }
)


@fixture
def resource_properties() -> ResourceProperties:
    return ResourceProperties(
        name="1",
        title="Test resource",
        description="A test resource",
        schema=TableSchemaProperties(
            fields=[
                FieldProperties(name="id", type="integer"),
                FieldProperties(name="name", type="string"),
            ]
        ),
    )


@fixture
def test_package(tmp_path):
    # TODO: Switch to using examples
    create_test_data_package(tmp_path)
    batch_path = tmp_path / "resources" / "1" / "batch"
    batch_path.mkdir(parents=True)

    for batch_data in [batch_data_1, batch_data_2]:
        batch_data.write_parquet(
            Path(batch_path / f"2025-03-26T100346Z-{uuid4()}").with_suffix(".parquet")
        )

    return tmp_path


@fixture
def resource_paths(test_package):
    return list((test_package / "resources" / "1" / "batch").iterdir())


def test_reads_resource_batches_correctly(resource_paths, resource_properties):
    """Reads the resource batches correctly with the expected timestamp column."""
    # Given, When
    data_list = read_resource_batches(
        resource_properties=resource_properties, paths=resource_paths
    )
    timestamp_column = [
        data_list[0][BATCH_TIMESTAMP_COLUMN_NAME],
        data_list[1][BATCH_TIMESTAMP_COLUMN_NAME],
    ]

    # Then
    assert len(data_list) == 2
    assert all(data.shape == (3, 3) for data in data_list)
    assert all(len(column.unique()) == 1 for column in timestamp_column)
    assert all(
        column.unique()[0] == "2025-03-26T100346Z" for column in timestamp_column
    )


def test_raises_error_when_file_does_not_exist(resource_paths, resource_properties):
    """Raises FileNotFoundError when a file in the list of paths doesn't exist"""
    # Given
    resource_paths.append(Path("non-existent-file.parquet"))

    # When, Then
    with raises(FileNotFoundError):
        read_resource_batches(
            resource_properties=resource_properties, paths=resource_paths
        )


def test_raises_error_when_file_not_parquet(tmp_path):
    """Raises a ValueError when the function is called with a non-Parquet file."""
    csv_file = tmp_path / "non-parquet-file.csv"
    csv_file.touch()

    with raises(ValueError):
        read_resource_batches(
            paths=[csv_file], resource_properties=example_resource_properties()
        )


def test_raises_error_when_timestamp_column_matches_existing_column(
    resource_paths, resource_properties
):
    """Raises ValueError when the timestamp column name matches an existing column."""
    # Given
    batch_path = resource_paths[0].parent

    batch_data = pl.DataFrame(
        {
            "id": [0, 1, 2],
            BATCH_TIMESTAMP_COLUMN_NAME: ["2024-03-26T100346Z"] * 3,
        }
    )
    batch_path = Path(batch_path) / f"2025-03-26T100346Z-{uuid4()}.parquet"
    batch_data.write_parquet(batch_path)

    # When, Then
    with raises(ValueError):
        read_resource_batches(
            resource_properties=resource_properties, paths=[batch_path]
        )


@mark.parametrize(
    "incorrect_timestamp",
    [
        "2025-55-26T100346Z",  # incorrect month
        "2025-02-30T100346",  # no timezone
        "2025-03-26T100346",  # incorrect date (30 February)
        "2025-03-26",  # no time
        "T100346Z",  # no date
    ],
)
def test_raises_error_when_file_name_timestamp_does_not_match_pattern(
    resource_paths, resource_properties, incorrect_timestamp
):
    """Raises ValueError when the batch file name is not in the expected pattern."""
    # Given
    batch_path = resource_paths[0].parent
    batch_file_path = Path(batch_path) / f"{incorrect_timestamp}-{uuid4()}.parquet"
    batch_data_1.write_parquet(batch_file_path)

    # When, Then
    with raises(ValueError):
        read_resource_batches(
            resource_properties=resource_properties, paths=[batch_file_path]
        )


def test_if_multiple_correct_timestamps_in_file_name_use_first_one(
    resource_paths, resource_properties
):
    """If multiple timestamps are found in the file name, the first one is used."""
    # Given
    batch_path = resource_paths[0].parent
    batch_file_path = (
        Path(batch_path) / f"2025-03-26T100346Z-1990-03-26T100346Z-{uuid4()}.parquet"
    )
    batch_data_1.write_parquet(batch_file_path)

    # When
    data_list = read_resource_batches(
        resource_properties=resource_properties, paths=[batch_file_path]
    )

    # Then
    assert data_list[0][BATCH_TIMESTAMP_COLUMN_NAME][0] == "2025-03-26T100346Z"


def test_raises_error_when_properties_do_not_match_data(
    resource_paths, resource_properties
):
    """Raises errors from checks when the resource properties don't match the data."""
    # Given
    resource_properties.schema.fields[0].name = "not-id"

    # When, Then
    with raises(ValueError):
        read_resource_batches(
            resource_properties=resource_properties, paths=resource_paths
        )


def test_raises_error_with_empty_resource_properties(resource_paths):
    """Raises errors from checks if the resource properties are empty."""
    # When, Then
    assert_raises_check_errors(
        lambda: read_resource_batches(
            resource_properties=ResourceProperties(), paths=resource_paths
        )
    )


def test_uses_cwd_if_no_paths(tmp_cwd, test_package, resource_properties):
    """If no paths are provided, should use the cwd as the package root to retrieve
    batch files from resource."""
    data_list = read_resource_batches(resource_properties)

    assert len(data_list) == 2


def test_no_error_thrown_when_no_batches_using_cwd_as_default(
    tmp_cwd, resource_properties
):
    """No error should be thrown if the cwd has no batch files for the given
    resource."""
    data_list = read_resource_batches(resource_properties)

    assert len(data_list) == 0



def get_json_path_to_resource_field(field: str, index: int | None = None) -> str:
    """Creates the JSON path to the specified field of a set of resource properties.

    Optionally adds the index of the resource properties, if they are part of a set of
    package properties.

    Args:
        field: The name of the field.
        index: The index of the resource properties. Defaults to None.

    Returns:
        The JSON path.
    """
    return "$." + ("" if index is None else f"resources[{index}].") + field



"""Mimicking the functional programming tools from R and the R package purrr."""

from itertools import repeat
from typing import Callable, TypeVar

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


def _map(x: list[T], fn: Callable[[T], U]) -> list[U]:
    """Use a function on each item in a list.

    This is a simpler, more user-friendly version of the built-in `map()`
    function. This function takes a list and a function, applies the function
    to each item in the list, and returns a new list with the results.
    This is mostly a thin wrapper around the built-in `map()` function.

    Inspired from the functionals from the R package purrr.

    Args:
        x: A list object.
        fn: A function to use on each item in the list `x`.

    Returns:
        A list with the results of applying `fn` to each item in `x`.
        The length of the list is the same as `x`.

    Examples:
        ```{python}
        from seedcase_sprout.internals import _map
        def square(x):
            return x * x
        _map([1, 2, 3], square)
        ```
    """
    return list(map(fn, x))


def _map2(x: list[T], y: list[U], fn: Callable[[T, U], V]) -> list[V]:
    """Use a function on each item in two lists, placed on the first two arguments.

    This function is similar to the `map2()` function from the R package purrr.
    It takes two lists and a function as arguments. The function is applied to
    each pair of elements from the two lists. If the second list has only one
    element, it is repeated for each element in the first list. The output
    is the same size as the first list.

    This makes the function a bit more user-friendly compared to the built-in
    `map()` function, which requires the same length for both lists.

    Args:
        x: The first list to use the function on.
        y: The second list that must have the same items as the first `x` list or one
            item to be repeated.
        fn: The function to use on each pair of items in the two lists above.

    Returns:
        A list of the results from the function used on each item in the two lists.

    Examples:
        ```{python}
        from seedcase_sprout.internals import _map2
        def add(a, b):
            return a + b
        _map2(
            x=[1, 2, 3],
            y=[10],
            fn=add
        )

        _map2(
            x=[1, 2, 3],
            y=[10, 20, 30],
            fn=add
        )
        ```
    """
    if len(y) == 1:
        y = list(repeat(y[0], len(x)))
    return list(map(fn, x, y))



from typing import Any

from seedcase_sprout.check_datapackage import CheckError, RequiredFieldType
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)

SPROUT_BLANK_ERROR_MESSAGE = "The '{field_name}' field is blank, please fill it in."


def check_fields_not_blank(
    properties: dict[str, Any],
    fields: dict[str, RequiredFieldType],
    index: int | None = None,
) -> list[CheckError]:
    """Checks that fields in `fields` are not blank if they are present.

    Fields not present in `properties` are not checked.

    For resource properties, an index may be supplied, if the resource properties are
    part of a set of package properties.

    Args:
        properties: The properties where the fields are.
        fields: A set of fields and their types.
        index: The index of the resource properties. Defaults to None.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    return [
        CheckError(
            message=SPROUT_BLANK_ERROR_MESSAGE.format(field_name=field),
            json_path=get_json_path_to_resource_field(field, index),
            validator="blank",
        )
        for field, type in fields.items()
        if properties.get(field) == get_blank_value_for_type(type)
    ]



from datetime import datetime
from typing import cast
from unittest.mock import patch
from uuid import UUID
from zoneinfo import ZoneInfo

import time_machine

from seedcase_sprout.create_properties_script import create_properties_script
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import LicenseProperties, PackageProperties
from tests.load_properties import load_properties


@patch("seedcase_sprout.properties.uuid4", return_value=UUID(int=1))
@time_machine.travel(datetime(2024, 5, 14, 5, 0, 1, tzinfo=ZoneInfo("UTC")), tick=False)
def test_creates_script_with_default_values(mock_uuid, tmp_cwd):
    """Should create a script with default values."""
    script_path = create_properties_script()

    assert script_path == PackagePath().properties_script()
    properties = load_properties(script_path, "package_properties")
    assert properties == PackageProperties(
        name=tmp_cwd.name,
        title="",
        description="",
        licenses=[LicenseProperties(name="")],
        id=str(mock_uuid()),
        version="0.1.0",
        created="2024-05-14T05:00:01+00:00",
    )


def test_works_with_custom_path(tmp_path):
    """Should work with a custom path."""
    script_path = create_properties_script(tmp_path)

    assert script_path == PackagePath(tmp_path).properties_script()
    properties = cast(
        PackageProperties, load_properties(script_path, "package_properties")
    )
    assert properties.name == tmp_path.name


def test_does_not_overwrite_existing_script(tmp_path):
    """Should not overwrite an existing script."""
    script_path = create_properties_script(tmp_path)

    # Create a file at the script path>
    script_path.write_text("This is a test file.")

    # Call the function again.
    new_script_path = create_properties_script(tmp_path)

    # The path should be the same, and the content should not change.
    assert new_script_path == script_path
    assert script_path.read_text() == "This is a test file."



"""This module contains functions to get the paths to various data package files.

They are intended to be used in conjunction with other functions to read, write, and
edit the contents and properties of various files within a data package. Specifically,
they are used in the context of a data package stored in the working directory ("local"
first approach).
"""

from pathlib import Path

from seedcase_sprout.internals import _create_resource_properties_script_filename


class PackagePath:
    """Get the absolute path to a specific file or folder in a data package.

    Use the functions in this class to get the absolute path to a specific file
    or folder in your data package. They are intended as convenience functions
    to provide you with easy and quick access to required files and folders
    within your data package.  These functions have these characteristics in
    common:

    -   All of these functions output a `Path` object.
    -   The base class has an optional `path` argument that defaults to
        the current working directory available from the base class.
    -   If the wrong `resource_name` is given, an error message will include a
        list of all the actual `resource_name`'s for a specific package.

    Outputs a `PackagePath` object representing the structure of a data package.

    Args:
        path: Provide a path to the package directory. Defaults to the current working
            directory.

    Examples:
        ```{python}
        from pathlib import Path

        import seedcase_sprout as sp

        # With default path
        sp.PackagePath().properties()
        sp.PackagePath().readme()

        # With custom path
        my_path = Path("my/path")
        sp.PackagePath(my_path).properties()
        sp.PackagePath(my_path).readme()
        ```
    """

    def __init__(self, path: Path | None = None):
        """Set the base path."""
        self.path = path or Path.cwd()

    def root(self) -> Path:
        """Path to the root folder of the package."""
        return self.path

    def properties(self) -> Path:
        """Path to the `datapackage.json` file."""
        return self.root() / "datapackage.json"

    def readme(self) -> Path:
        """Path to the `README.md` file."""
        return self.root() / "README.md"

    def resources(self) -> Path:
        """Path to the `resources/` folder."""
        return self.root() / "resources"

    def resource(self, resource_name: str) -> Path:
        """Path to the specified `resources/<name>/` folder.

        Args:
            resource_name: The name of the resource. Use `ResourceProperties.name` to
                get the correct resource name.
        """
        return self.resources() / str(resource_name)

    def resource_data(self, resource_name: str) -> Path:
        """Path to the specific resource's data file.

        Args:
            resource_name: The name of the resource. Use `ResourceProperties.name` to
                get the correct resource name.
        """
        return self.resource(resource_name) / "data.parquet"

    def resource_batch(self, resource_name: str) -> Path:
        """Path to the specific resource's `batch/` folder.

        Args:
            resource_name: The name of the resource. Use `ResourceProperties.name` to
                get the correct resource name.
        """
        return self.resource(resource_name) / "batch"

    def resource_batch_files(self, resource_name: str) -> list[Path]:
        """Paths to all the files in the specific resource's `batch/` folder.

        Args:
            resource_name: The name of the resource. Use `ResourceProperties.name` to
                get the correct resource name.
        """
        return list(self.resource_batch(resource_name).glob("*.parquet"))

    def properties_script(self) -> Path:
        """Path to the properties script."""
        return self.root() / "scripts" / "package_properties.py"

    def resource_properties_script(self, resource_name: str = "") -> Path:
        """Path to a specific resource's resource properties script.

        Args:
            resource_name: The name of the resource.
        """
        return (
            self.root()
            / "scripts"
            / f"{_create_resource_properties_script_filename(resource_name)}.py"
        )



from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)


def test_returns_expected_json_path_without_index():
    """Should form the correct JSON path with no index supplied."""
    assert get_json_path_to_resource_field("myField") == "$.myField"


def test_returns_correct_path_with_index():
    """Should form the correct JSON path with a resource index supplied."""
    assert get_json_path_to_resource_field("myField", 2) == "$.resources[2].myField"



from pytest import mark

from seedcase_sprout.check_datapackage import RequiredFieldType
from seedcase_sprout.sprout_checks.check_fields_present import (
    check_fields_present,
)
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)

REQUIRED_FIELDS = {"name": RequiredFieldType.str, "tags": RequiredFieldType.list}


@mark.parametrize("index", [None, 2])
def test_no_error_found_in_properties_with_required_fields(index):
    """Should pass properties with required fields present and populated."""
    properties = {"name": "My name", "tags": ["a", "b"]}

    assert check_fields_present(properties, REQUIRED_FIELDS, index) == []


@mark.parametrize("index", [None, 2])
def test_no_error_found_in_properties_with_required_fields_blank(index):
    """Should pass properties with required fields present but blank."""
    properties = {"name": "", "tags": []}

    assert check_fields_present(properties, REQUIRED_FIELDS, index) == []


@mark.parametrize("index", [None, 2])
def test_error_found_if_there_is_a_missing_required_field(index):
    """Should find an error if there is a missing required field."""
    properties = {"name": "My name"}

    errors = check_fields_present(properties, REQUIRED_FIELDS, index)

    assert len(errors) == 1
    assert "required" in errors[0].message
    assert errors[0].json_path == get_json_path_to_resource_field("tags", index)
    assert errors[0].validator == "required"


@mark.parametrize("index", [None, 2])
def test_error_found_if_there_are_multiple_missing_required_fields(index):
    """Should find an error if there are multiple missing required fields."""
    errors = check_fields_present({}, REQUIRED_FIELDS, index)

    assert len(errors) == 2
    assert all(error.validator == "required" for error in errors)
    assert any(
        error.json_path == get_json_path_to_resource_field("name", index)
        for error in errors
    )
    assert any(
        error.json_path == get_json_path_to_resource_field("tags", index)
        for error in errors
    )



from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path

from seedcase_sprout.properties import Properties


def load_properties(path: Path, object_name: str) -> Properties:
    """Loads `Properties` object from file."""
    spec = spec_from_file_location("test_module", path)
    assert spec
    assert spec.loader
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    properties: Properties = getattr(module, object_name)
    return properties



from pytest import mark

from seedcase_sprout.check_datapackage import RequiredFieldType
from seedcase_sprout.sprout_checks.check_list_item_field_not_blank import (
    check_list_item_field_not_blank,
)
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)


def test_no_error_found_in_properties_without_list():
    """Should pass if the properties do not contain the specified list."""
    assert check_list_item_field_not_blank({}, "items", "field") == []


@mark.parametrize("items", [[], [{}, {}], [{"a": 1}, {"a": 2}]])
def test_no_error_found_when_list_does_not_contain_field(items):
    """Should pass if list items do not contain the field."""
    properties = {"items": items}

    assert check_list_item_field_not_blank(properties, "items", "field") == []


def test_no_error_found_when_fields_populated():
    """Should pass if all fields are populated."""
    properties = {"items": [{"field": "value"}, {"field": "value"}]}

    assert check_list_item_field_not_blank(properties, "items", "field") == []


def test_no_error_found_when_fields_are_of_wrong_type():
    """Should pass if the fields are present but of the wrong type."""
    properties = {"items": [{"field": "value"}, {"field": ""}]}

    assert (
        check_list_item_field_not_blank(
            properties, "items", "field", RequiredFieldType.list
        )
        == []
    )


@mark.parametrize(
    "field_type,value",
    [(RequiredFieldType.str, "value"), (RequiredFieldType.list, [1])],
)
def test_error_found_if_an_item_has_a_blank_field(field_type, value):
    """Should find an error if there is an item with a blank field."""
    properties = {
        "items": [{"field": value}, {"field": get_blank_value_for_type(field_type)}]
    }

    errors = check_list_item_field_not_blank(properties, "items", "field", field_type)

    assert len(errors) == 1
    assert "blank" in errors[0].message
    assert errors[0].json_path == "$.items[1].field"
    assert errors[0].validator == "blank"


@mark.parametrize("field_type", RequiredFieldType)
def test_error_found_if_multiple_items_have_a_blank_field(field_type):
    """Should find an error if there are multiple items with a blank field."""
    properties = {"items": [{"field": get_blank_value_for_type(field_type)}] * 2}

    errors = check_list_item_field_not_blank(properties, "items", "field", field_type)

    assert len(errors) == 2
    assert all(
        "blank" in error.message and error.validator == "blank" for error in errors
    )
    assert any(error.json_path == "$.items[0].field" for error in errors)
    assert any(error.json_path == "$.items[1].field" for error in errors)



from functools import total_ordering


@total_ordering
class CheckError(Exception):
    """Raised or returned when a properties object fails a single check."""

    def __init__(
        self,
        message: str,
        json_path: str,
        validator: str,
    ):
        """Initialises CheckError.

        Args:
            message: The error message.
            json_path: The path to the JSON field within the enclosing JSON object where
                the error occurred.
            validator: The name of the validator that failed.
        """
        self.message = message
        self.json_path = json_path
        self.validator = validator
        super().__init__(self.__str__())

    def __eq__(self, other: object) -> bool:
        """Checks if this error is equal to an object.

        Args:
            other: The object to compare.

        Returns:
            If the objects are equal.
        """
        if not isinstance(other, CheckError):
            return NotImplemented
        return (
            self.message == other.message
            and self.json_path == other.json_path
            and self.validator == other.validator
        )

    def __lt__(self, other: object) -> bool:
        """Checks if this error is less than an object.

        Args:
            other: The object to compare.

        Returns:
            The result of the comparison.
        """
        if not isinstance(other, CheckError):
            return NotImplemented
        return (self.json_path, self.validator, self.message) < (
            other.json_path,
            other.validator,
            other.message,
        )

    def __hash__(self) -> int:
        """Returns a hash for this error.

        Returns:
            The hash.
        """
        return hash((self.message, self.json_path, self.validator))

    def __str__(self) -> str:
        """Returns a user-friendly string representation of the error.

        Returns:
            The string representation.
        """
        return (
            f"Error at `{self.json_path}` caused by `{self.validator}`: {self.message}"
        )

    def __repr__(self) -> str:
        """Returns a developer-friendly, unambiguous representation of the error.

        Returns:
            The developer-friendly representation.
        """
        return (
            f"CheckError(message={self.message!r}, "
            f"json_path={self.json_path!r}, "
            f"validator={self.validator!r})"
        )



"""Check functions and constants for the Frictionless Data Package standard."""

from .check_error import CheckError
from .check_error_matcher import CheckErrorMatcher
from .check_package_properties import check_package_properties
from .check_properties import check_properties
from .check_resource_properties import check_resource_properties
from .constants import (
    PACKAGE_RECOMMENDED_FIELDS,
    PACKAGE_REQUIRED_FIELDS,
    RESOURCE_REQUIRED_FIELDS,
    RequiredFieldType,
)
from .exclude_matching_errors import exclude_matching_errors

__all__ = [
    "CheckError",
    "CheckErrorMatcher",
    "check_properties",
    "check_package_properties",
    "check_resource_properties",
    "PACKAGE_RECOMMENDED_FIELDS",
    "PACKAGE_REQUIRED_FIELDS",
    "RESOURCE_REQUIRED_FIELDS",
    "RequiredFieldType",
    "exclude_matching_errors",
]



from typing import cast

import polars as pl

from seedcase_sprout.check_properties import (
    check_resource_properties,
)
from seedcase_sprout.internals import _get_nested_attr, _map
from seedcase_sprout.map_data_types import (
    _get_allowed_polars_types,
    _polars_and_datapackage_types_match,
)
from seedcase_sprout.properties import (
    FieldProperties,
    ResourceProperties,
)


def check_data(
    data: pl.DataFrame, resource_properties: ResourceProperties
) -> pl.DataFrame:
    """Check that the DataFrame matches the requirements in the resource properties.

    Run a few checks to compare between the data and the properties on the items:

    | Data | Properties |
    |:------|:------------|
    | Column names | `field.name` |
    | Column types | `field.types` |
    | Column values' types | `field.types` |
    | Column values' constraints | `field.constraints` |

    The error messages are generally in the format of:

    ```
    # {data item}:

    There is a mismatch found:

    - In the properties: {mismatch}
    - In the data: {mismatch}
    ```

    Args:
        data: A Polars DataFrame.
        resource_properties: The specific `ResourceProperties` for the `data`.

    Returns:
        The `data` if checks all pass.

    Raises:
        ExceptionGroup[CheckError]: If the resource properties are incorrect.
        ValueError: If column names in the data are incorrect.
        ExceptionGroup[ValueError]: If data types in the data are incorrect.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        sp.check_data(
            data=sp.example_data(),
            resource_properties=sp.example_resource_properties()
        )
        ```
    """
    check_resource_properties(resource_properties)
    _check_column_names(data, resource_properties)
    _check_column_types(data, resource_properties)
    # _check_column_values_constraints(data, resource_properties)

    return data


def _check_column_names(
    data: pl.DataFrame, resource_properties: ResourceProperties
) -> pl.DataFrame:
    """Checks that column names in `data` match those in `resource_properties`.

    Columns may appear in any order.

    Args:
        data: The data to check.
        resource_properties: The resource properties to check against.

    Returns:
        The data if the column names match.

    Raises:
        ValueError: If the column names don't match the names in
            `resource_properties`.
    """
    columns_in_data = data.schema.names()
    columns_in_resource = [
        str(field.name)
        for field in cast(
            list[FieldProperties],
            _get_nested_attr(resource_properties, "schema.fields", default=[]),
        )
    ]
    extra_columns_in_data = [
        name for name in columns_in_data if name not in columns_in_resource
    ]
    missing_columns_in_data = [
        name for name in columns_in_resource if name not in columns_in_data
    ]

    if extra_columns_in_data or missing_columns_in_data:
        raise ValueError(
            _format_column_name_error_message(
                extra_columns_in_data, missing_columns_in_data
            )
        )

    return data


def _format_column_name_error_message(
    extra_columns_in_data: list[str], missing_columns_in_data: list[str]
) -> str:
    message = (
        "Column names in the data do not match column names in the resource properties:"
    )
    if extra_columns_in_data:
        message += f"\n- Unexpected column(s) in data: {extra_columns_in_data}"
    if missing_columns_in_data:
        message += f"\n- Missing column(s) in data: {missing_columns_in_data}"
    return message


def _check_column_types(
    data: pl.DataFrame, resource_properties: ResourceProperties
) -> pl.DataFrame:
    """Checks that column data types match the data types specified in the properties.

    The resource properties specify a Frictionless data type for each column.
    This function checks if the Polars data type of each column in the data matches
    the expected Frictionless data type.

    Column names are expected to match the names specified in the resource properties.

    Args:
        data: The data frame to check.
        resource_properties: The resource properties to check against.

    Returns:
        The data frame, if all column types are correct.

    Raises:
        ExceptionGroup: A group of `ValueError`s, one per incorrectly typed column.
    """
    fields = cast(
        list[FieldProperties],
        _get_nested_attr(resource_properties, "schema.fields", default=[]),
    )
    polars_schema = data.schema
    errors = [
        _get_column_type_error(polars_schema[str(field.name)], field)
        for field in fields
        if not _polars_and_datapackage_types_match(
            polars_schema[str(field.name)], field.type
        )
    ]

    if errors:
        raise ExceptionGroup(
            (
                "The following columns in the data have data types that do not match "
                "the data types in the resource properties:"
            ),
            errors,
        )
    return data


def _get_column_type_error(
    polars_type: pl.DataType, field: FieldProperties
) -> ValueError:
    """Creates an error for a column where Polars and Frictionless types don't match.

    Args:
        polars_type: The Polars type.
        field: The field properties.

    Returns:
        A `ValueError`.
    """
    allowed_types = _map(_get_allowed_polars_types(field.type), str)
    allowed_types_str = (
        allowed_types[0]
        if len(allowed_types) == 1
        else f"one of {', '.join(allowed_types)}"
    )

    if field.type == "geopoint":
        allowed_types_str = "an Array of a numeric type with size 2"

    return ValueError(
        f"Expected type of column '{field.name}' "
        f"to be {allowed_types_str} but found {polars_type}."
    )



from pytest import fixture, mark

from seedcase_sprout.sprout_checks.check_required_package_properties_not_blank import (  # noqa: E501
    check_required_package_properties_not_blank,
)
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)
from seedcase_sprout.sprout_checks.required_fields import (
    PACKAGE_SPROUT_REQUIRED_FIELDS,
)


@fixture
def properties():
    return {
        "name": "package-1",
        "id": "abc1",
        "title": "Package 1",
        "description": "A package.",
        "version": "1.0.0",
        "created": "2024-05-14T05:00:01+00:00",
        "licenses": [{"name": "a-license"}],
        "contributors": [{"title": "a contributor"}],
        "sources": [{"title": "a source"}],
    }


def test_no_error_found_if_all_required_fields_populated(properties):
    """Should pass if all required fields are present and populated."""
    assert check_required_package_properties_not_blank(properties) == []


def test_no_error_found_if_all_required_fields_missing():
    """Should pass if all required fields are missing."""
    assert check_required_package_properties_not_blank({}) == []


@mark.parametrize(
    "field,value,json_path",
    [
        *[
            (name, get_blank_value_for_type(type), f"$.{name}")
            for name, type in PACKAGE_SPROUT_REQUIRED_FIELDS.items()
        ],
        ("contributors", [{"title": ""}], "$.contributors[0].title"),
        ("sources", [{"title": ""}], "$.sources[0].title"),
        ("licenses", [{"name": ""}], "$.licenses[0].name"),
        ("licenses", [{"path": ""}], "$.licenses[0].path"),
    ],
)
def test_error_found_if_required_field_is_blank(properties, field, value, json_path):
    """Should find an error if a required field is present but blank."""
    properties[field] = value

    errors = check_required_package_properties_not_blank(properties)

    assert len(errors) == 1
    assert errors[0].json_path == json_path
    assert errors[0].validator == "blank"


def test_error_found_if_all_required_fields_are_blank():
    """Should find an error if all required fields are present but blank."""
    properties = {
        "name": "",
        "id": "",
        "title": "",
        "description": "",
        "version": "",
        "created": "",
        "licenses": [{"name": "", "path": ""}],
        "contributors": [{"title": ""}],
        "sources": [{"title": ""}],
    }

    errors = check_required_package_properties_not_blank(properties)

    assert len(errors) == 10
    assert all(error.validator == "blank" for error in errors)



from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.check_error_matcher import CheckErrorMatcher
from seedcase_sprout.check_datapackage.exclude_matching_errors import (
    exclude_matching_errors,
)

errors = [
    CheckError("'path' is a required property", "$.path", "required"),
    CheckError("'name' is a required property", "$.name", "required"),
    CheckError("123 is not of type 'string'", "$.resources[0].name", "type"),
    CheckError("pattern 'xyz' doesn't match", "$.created", "pattern"),
    CheckError("pattern 'xyz' doesn't match", "$.version", "pattern"),
]


def test_empty_matchers_have_no_effect():
    """An empty list as a list of matchers should have no effect."""
    assert exclude_matching_errors(errors, []) == errors


def test_not_matching_matchers_have_no_effect():
    """If no matchers match, no errors should be excluded."""
    assert (
        exclude_matching_errors(
            errors,
            [
                CheckErrorMatcher(validator="no-match"),
                CheckErrorMatcher(
                    validator="required", json_path="path", message="no match!"
                ),
                CheckErrorMatcher(
                    validator="type", json_path=r"\$no\.match", message="123 is not"
                ),
            ],
        )
        == errors
    )


def test_matched_errors_are_excluded():
    """If any matchers match, the error should be excluded."""
    assert (
        exclude_matching_errors(
            errors,
            [
                CheckErrorMatcher(json_path="name", validator="required"),
                CheckErrorMatcher(validator="pattern"),
                CheckErrorMatcher(validator="type"),
                CheckErrorMatcher(message="type 'string'"),
            ],
        )
        == errors[:1]
    )



from typing import Any

import seedcase_sprout.check_datapackage as cdp
from seedcase_sprout.properties import PackageProperties, ResourceProperties
from seedcase_sprout.sprout_checks.get_sprout_package_errors import (
    get_sprout_package_errors,
)
from seedcase_sprout.sprout_checks.get_sprout_resource_errors import (
    get_sprout_resource_errors,
)

# Script only constant
_RESOURCE_FIELD_PATTERN = r"resources\[\d+\]"


def check_package_properties(properties: Any) -> PackageProperties:
    """Check `PackageProperties` (not `ResourceProperties`) against the requirements.

    Package `properties` are checked against the Data Package standard and the following
    Sprout-specific requirements:

    - Sprout-specific required fields are present.
    - Required fields are not blank.

    Args:
        properties: The package properties to check.

    Returns:
        The `properties` if all checks pass.

    Raises:
        ExceptionGroup: A group of `CheckError`s, one error per failed check.
    """
    package_properties = _check_is_package_properties_type(properties)
    return _generic_check_properties(
        package_properties,
        ignore=[
            # Ignore checks on specific resources within the resource field.
            cdp.CheckErrorMatcher(json_path=_RESOURCE_FIELD_PATTERN),
            # Ignore missing resources.
            cdp.CheckErrorMatcher(json_path=r"resources$", validator="required"),
        ],
    )


def check_properties(properties: Any) -> PackageProperties:
    """Check that all `properties` match Sprout's requirements.

    If the resources property hasn't been filled in yet, this will only check
    the package properties. The `properties` are checked against the Data
    Package standard and the following Sprout-specific requirements:

    - Sprout-specific required fields are present
    - Required fields are not blank

    If the resources property *has* been filled in, these resource properties will also
    be checked:

    - `path` is of type string.
    - `path` includes resource name.
    - `data` is not set.

    Args:
        properties: The properties to check.

    Returns:
        The `properties` if all checks pass.

    Raises:
        ExceptionGroup: A group of `CheckError`s, one error per failed check.
    """
    package_properties = _check_is_package_properties_type(properties)
    if not package_properties.resources:
        check_package_properties(package_properties)
    else:
        _generic_check_properties(package_properties)
    return package_properties


def check_resource_properties(properties: Any) -> ResourceProperties:
    """Checks that only the resource `properties` match Sprout's requirements.

    All resource `properties` are checked against the Data Package standard and
    the following Sprout-specific requirements:

    - Sprout-specific required fields are present.
    - Required fields are not blank.
    - `path` is of type string.
    - `path` includes resource name.
    - `data` is not set.

    Args:
        properties: The resource properties to check.

    Returns:
        Outputs the `properties` if all checks pass.

    Raises:
        ExceptionGroup: A group of `CheckError`s, one error per failed check.
    """
    package_field_pattern = r"\$\.\w+$"
    resource_properties = _check_is_resource_properties_type(properties)
    try:
        _generic_check_properties(
            PackageProperties(resources=[resource_properties]),
            ignore=[cdp.CheckErrorMatcher(json_path=package_field_pattern)],
        )
    # TODO: This probably is better placed in the `check-datapackage` package.
    except ExceptionGroup as error_info:
        for error in error_info.exceptions:
            if isinstance(error, cdp.CheckError):
                error.json_path = error.json_path.replace(".resources[0]", "")
        raise error_info

    return resource_properties


def _generic_check_properties(
    properties: PackageProperties, ignore: list[cdp.CheckErrorMatcher] = []
) -> PackageProperties:
    """A generic check for Sprout-specific requirements on the Frictionless standard.

    All `properties`, excluding those in `ignore`, are checked against the Data
    Package standard as well as the following Sprout-specific requirements:

    - Sprout-specific required fields are present
    - Required fields are not blank
    - Resource `path` is of type string
    - Resource `path` includes resource name
    - Resource `data` is not set

    Args:
        properties: The full package properties to check, including resource properties.
        ignore: A list of matchers for any `CheckErrors` to ignore.

    Returns:
        Outputs the `properties` object if all checks pass.

    Raises:
        ExceptionGroup: A group of `CheckError`s, one error per failed check.
    """
    properties_dict = properties.compact_dict

    errors = cdp.check_properties(properties_dict)

    errors += get_sprout_package_errors(properties_dict)

    resources = properties_dict.get("resources")
    if isinstance(resources, list):
        for index, resource in enumerate(resources):
            if isinstance(resource, dict):
                errors += get_sprout_resource_errors(resource, index)

    errors = cdp.exclude_matching_errors(
        errors,
        [
            *ignore,
            cdp.CheckErrorMatcher(
                validator="required", json_path=rf"{_RESOURCE_FIELD_PATTERN}\.data$"
            ),
            cdp.CheckErrorMatcher(
                validator="type",
                json_path=rf"{_RESOURCE_FIELD_PATTERN}\.path$",
                message="not of type 'array'",
            ),
        ],
    )
    errors = sorted(set(errors))

    if errors:
        raise ExceptionGroup(
            f"The following checks failed on the properties:\n{properties}", errors
        )

    return properties


def _check_is_package_properties_type(properties: Any) -> PackageProperties:
    if not isinstance(properties, PackageProperties):
        raise TypeError(
            f"Expected properties to be a PackageProperties object,"
            f"but the object is {type(properties)}"
        )
    return properties


def _check_is_resource_properties_type(properties: Any) -> ResourceProperties:
    if not isinstance(properties, ResourceProperties):
        raise TypeError(
            f"Expected properties to be a ResourceProperties object,"
            f"but the object is {type(properties)}"
        )
    return properties



from datetime import datetime

from jinja2 import Environment, FileSystemLoader

from seedcase_sprout.constants import TEMPLATES_PATH
from seedcase_sprout.internals import _to_dedented
from seedcase_sprout.properties import (
    LicenseProperties,
    PackageProperties,
)


def as_readme_text(properties: PackageProperties) -> str:
    """Create a human-readable version of the properties as a README string.

    Convert the properties object into human-readable version as Markdown-formatted
    text. Use this to generate the text for a `README.md` file and save it using
    `write_file()`.

    Args:
        properties: A `PackageProperties` object containing the package and resource
            properties.

    Returns:
        A Markdown-styled string to eventually save as a `README.md` file.
    """
    env = Environment(loader=FileSystemLoader(TEMPLATES_PATH), autoescape=True)
    env.filters["join_names"] = _join_names
    env.filters["format_date"] = _format_date
    env.filters["inline_code"] = _inline_code
    env.filters["format_link"] = _format_link
    template = env.get_template("README.jinja2")

    # Dedent description text.
    if properties.description:
        properties.description = _to_dedented(properties.description)

    return template.render(properties=properties)


def _join_names(licenses: list[LicenseProperties] | None) -> str:
    """Joins license names into a comma-separated list.

    Args:
        licenses: The licenses.

    Returns:
        A comma-separated list of names.
    """
    return ", ".join(str(license.name) for license in licenses) if licenses else "N/A"


def _format_date(created: str | None) -> str:
    """Transforms ISO date stamp to human-readable format.

    Args:
        created: The ISO date stamp.

    Returns:
        The date in a human-readable format.
    """
    return (
        datetime.fromisoformat(created).strftime("%d %B %Y, %H:%M")
        if created
        else "N/A"
    )


def _inline_code(value: str | None) -> str:
    """Adds inline code formatting to the input.

    Args:
        value: The value to format as inline code.

    Returns:
        The value formatted as inline code.
    """
    return f"`{value}`" if value else "N/A"


def _format_link(url: str | None, title: str = "See more") -> str:
    """Formats a URL as a markdown link.

    Args:
        url: The URL to format.
        title: The title to show for the link. Defaults to "See more".

    Returns:
        A markdown link using the URL and the title.
    """
    return f"[{title}]({url})" if url else "N/A"



import seedcase_sprout.check_datapackage as cdp

# Sprout-specific required fields and their types

PACKAGE_SPROUT_REQUIRED_FIELDS = (
    cdp.PACKAGE_REQUIRED_FIELDS
    | cdp.PACKAGE_RECOMMENDED_FIELDS
    | {
        "title": cdp.RequiredFieldType.str,
        "description": cdp.RequiredFieldType.str,
        "version": cdp.RequiredFieldType.str,
        "created": cdp.RequiredFieldType.str,
    }
)
PACKAGE_SPROUT_REQUIRED_FIELDS.pop("resources", None)

RESOURCE_SPROUT_REQUIRED_FIELDS = cdp.RESOURCE_REQUIRED_FIELDS | {
    "title": cdp.RequiredFieldType.str,
    "description": cdp.RequiredFieldType.str,
}
RESOURCE_SPROUT_REQUIRED_FIELDS.pop("data", None)



from pytest import fixture, raises

from seedcase_sprout.write_file import write_file


@fixture
def file_content():
    return "This is test content."


def test_returns_path(tmp_path, file_content):
    """Tests that the path is returned."""
    file_path = tmp_path / "test.txt"

    assert write_file(file_content, file_path) == file_path


def test_creates_a_file_with_the_given_content(tmp_path, file_content):
    """Tests that a file with the given content is created."""
    file_path = tmp_path / "test.md"

    assert write_file(file_content, file_path) == file_path
    assert file_path.is_file()
    assert file_path.read_text() == file_content


def test_raises_error_when_parent_folder_does_not_exist(tmp_path, file_content):
    """Tests that FileNotFoundError is raised when the parent folder doesn't exist."""
    file_path = tmp_path / "non_existent_folder" / "test.md"

    with raises(FileNotFoundError):
        write_file(file_content, file_path)


def test_overwrites_file_content_if_file_already_exists(tmp_path, file_content):
    """Tests that if file already exists, it will be overwritten."""
    file_path = tmp_path / "test.txt"
    write_file(file_content, file_path)
    new_file_content = "This is new content."

    assert write_file(new_file_content, file_path) == file_path
    assert file_path.read_text() == new_file_content



from typing import Any

from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.constants import DATA_PACKAGE_SCHEMA_PATH
from seedcase_sprout.check_datapackage.internals import (
    _add_package_recommendations,
    _add_resource_recommendations,
    _check_object_against_json_schema,
    _read_json,
)


def check_properties(
    properties: dict[str, Any], check_recommendations: bool = True
) -> list[CheckError]:
    """Checks that `properties` matches the Data Package standard (v2.0).

    Both package and resource properties are checked. Structural, type and format
    constraints are all checked. All schema violations are
    collected before errors are returned.

    The schema loaded or constructed in this function overrides any values specified
    in the `$schema` attribute of `properties`, including the default value.

    Args:
        properties: The full package properties to check, including resource properties.
        check_recommendations: Whether `properties` should be checked against
            recommendations in the Data Package standard in addition to requirements.
            Defaults to True.

    Returns:
        A list of errors. An empty list, if no errors are found.
    """
    schema = _read_json(DATA_PACKAGE_SCHEMA_PATH)

    if check_recommendations:
        _add_package_recommendations(schema)
        _add_resource_recommendations(schema)

    return _check_object_against_json_schema(properties, schema)



from pytest import raises

from seedcase_sprout import ExamplePackage, example_data, read_resource_data


def test_read_resource_data():
    """Very basic check that it runs."""
    with ExamplePackage():
        data = read_resource_data("example-resource")
        assert data.shape == example_data().shape


def test_read_resource_data_with_package_path():
    """Check how it works when given a path."""
    with ExamplePackage() as package_path:
        data = read_resource_data("example-resource", package_path.root())
        assert data.shape == example_data().shape


def test_read_resource_data_with_wrong_resource():
    with raises(FileNotFoundError):
        read_resource_data("wrong-name")



from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.check_error_matcher import CheckErrorMatcher


def exclude_matching_errors(
    errors: list[CheckError], matchers: list[CheckErrorMatcher]
) -> list[CheckError]:
    """Returns a new list of errors, with errors matched by any `matchers` filtered out.

    Args:
        errors: The errors to exclude matching errors from.
        matchers: The matches to exclude.

    Returns:
        A list of errors without any errors that matched.
    """
    return [
        error
        for error in errors
        if not any(matcher.matches(error) for matcher in matchers)
    ]



import re

import polars as pl
from polars.testing import assert_frame_equal
from pytest import fixture, mark, raises

from seedcase_sprout.check_data import check_data
from seedcase_sprout.examples import (
    example_data,
    example_resource_properties,
)
from seedcase_sprout.map_data_types import _get_allowed_polars_types
from seedcase_sprout.properties import (
    FieldProperties,
    ResourceProperties,
    TableSchemaProperties,
)
from tests.assert_raises_errors import (
    assert_raises_check_errors,
    assert_raises_errors,
)

string_field = FieldProperties(name="my_string", type="string")
bool_field = FieldProperties(name="my_bool", type="boolean")
number_field = FieldProperties(name="my_number", type="number")


@fixture
def resource_properties() -> ResourceProperties:
    return ResourceProperties(
        name="data",
        title="data",
        description="My data...",
        schema=TableSchemaProperties(),
    )


@mark.parametrize(
    "fr_type, pl_types",
    [
        ("number", [pl.Float32, pl.Float64, pl.Decimal]),
        ("integer", [pl.Int16, pl.UInt8, pl.Duration]),
        ("string", [pl.String, pl.Categorical, pl.Enum(["A"]), pl.Binary]),
        ("date", [pl.Date]),
        ("datetime", [pl.Datetime]),
        ("time", [pl.Time]),
        ("boolean", [pl.Boolean]),
        ("array", [pl.String, pl.Array, pl.List]),
        ("object", [pl.String, pl.Struct, pl.Object]),
        (
            "geopoint",
            [pl.Array(pl.Int128, 2), pl.Array(pl.Float32, 2), pl.Array(pl.Decimal, 2)],
        ),
        ("geojson", [pl.String, pl.Struct, pl.Object]),
        ("year", [pl.Int32, pl.UInt8, pl.Int64]),
        ("yearmonth", [pl.Date]),
        ("duration", [pl.String]),
        ("any", [pl.Unknown, pl.Null, pl.Object, pl.Int128, pl.Boolean]),
    ],
)
def test_accepts_correct_columns(fr_type, pl_types):
    """Should not raise an error when columns in the data match columns in the
    resource properties."""
    resource_properties = example_resource_properties()
    assert resource_properties.schema
    resource_properties.schema.fields = [
        FieldProperties(name=str(pl_type), type=fr_type) for pl_type in pl_types
    ]
    data = pl.DataFrame(
        {str(pl_type): pl.Series([], dtype=pl_type) for pl_type in pl_types}
    )

    assert check_data(data, resource_properties) is data


def test_accepts_columns_in_any_order():
    """Should not raise an error when the data types match but the columns are in a
    different order."""
    resource_properties = example_resource_properties()
    assert resource_properties.schema and resource_properties.schema.fields
    resource_properties.schema.fields.reverse()
    data = example_data()

    assert_frame_equal(check_data(data, resource_properties), data)


def test_throws_error_when_data_has_extra_columns(resource_properties):
    """Should throw an error if the data has extra columns."""
    df = pl.DataFrame({"extra_col1": [""], bool_field.name: [True], "extra_col2": [""]})
    resource_properties.schema.fields = [bool_field]

    with raises(ValueError, match=r"Unexpected.*extra_col1.*extra_col2") as error:
        check_data(df, resource_properties)

    assert bool_field.name and bool_field.name not in str(error)


def test_throws_error_when_data_has_missing_columns(resource_properties):
    """Should throw an error if the data has missing columns."""
    df = pl.DataFrame({bool_field.name: [True]})
    resource_properties.schema.fields = [string_field, bool_field, number_field]

    with raises(
        ValueError, match=rf"Missing.*{string_field.name}.*{number_field.name}"
    ) as error:
        check_data(df, resource_properties)

    assert bool_field.name and bool_field.name not in str(error)


def test_throws_error_when_data_has_extra_and_missing_columns(resource_properties):
    """Should throw an error if the data has extra columns and missing columns."""
    df = pl.DataFrame({"extra_col": [""], bool_field.name: [True]})
    resource_properties.schema.fields = [string_field, bool_field, number_field]

    with raises(ValueError) as error:
        check_data(df, resource_properties)

    assert re.search(r"Unexpected.*extra_col", str(error))
    assert re.search(rf"Missing.*{string_field.name}.*{number_field.name}", str(error))
    assert bool_field.name and bool_field.name not in str(error)


@mark.parametrize(
    "frictionless_type",
    [
        "string",
        "boolean",
        "integer",
        "number",
        "year",
        "datetime",
        "date",
        "time",
        "yearmonth",
        "geopoint",
        "duration",
        "object",
        "array",
        "geojson",
    ],
)
def test_rejects_incorrect_column_type(frictionless_type, resource_properties):
    """Should raise an error if the Polars type does not match the Frictionless type."""
    data = pl.DataFrame({"my_col": pl.Series([], dtype=pl.Null)})
    resource_properties.schema.fields = [
        FieldProperties(name="my_col", type=frictionless_type)
    ]

    with raises(ExceptionGroup) as error_info:
        check_data(data, resource_properties)

    errors = error_info.value.exceptions
    assert len(errors) == 1
    polars_type = re.escape(str(data.schema.dtypes()[0]))
    allowed_types = _get_allowed_polars_types(frictionless_type)
    assert re.search(rf"'my_col'.*{allowed_types}.*found {polars_type}", str(errors[0]))


def test_rejects_multiple_incorrect_column_types():
    """Should raise an error if multiple column types don't match."""
    resource_properties = example_resource_properties()
    data = example_data()
    data = data.select([pl.lit(None).alias(col) for col in data.columns])

    with raises(ExceptionGroup) as error_info:
        check_data(data, resource_properties)

    errors = error_info.value.exceptions
    assert len(errors) == data.width
    assert resource_properties.schema and resource_properties.schema.fields
    for error, field in zip(errors, resource_properties.schema.fields):
        polars_type = re.escape(str(data.schema[str(field.name)]))
        allowed_types = _get_allowed_polars_types(field.type)
        assert re.search(
            rf"'{field.name}'.*{allowed_types}.*found {polars_type}", str(error)
        )


def test_rejects_geopoint_with_incorrect_size():
    """Should raise an error if the size of the array representing a geopoint is not
    correct."""
    data = pl.DataFrame(
        {"my_geopoint": pl.Series([[1, 1, 1]] * 3, dtype=pl.Array(pl.Int8, 3))}
    )
    resource_properties = example_resource_properties()
    assert resource_properties.schema
    resource_properties.schema.fields = [
        FieldProperties(name="my_geopoint", type="geopoint")
    ]

    assert_raises_errors(lambda: check_data(data, resource_properties), ValueError, 1)


def test_rejects_geopoint_with_incorrect_inner_type():
    """Should raise an error if the type of the nested elements in a geopoint array is
    not correct."""
    data = pl.DataFrame(
        {"my_geopoint": pl.Series([["a", "b"]] * 3, dtype=pl.Array(pl.String, 2))}
    )
    resource_properties = example_resource_properties()
    assert resource_properties.schema
    resource_properties.schema.fields = [
        FieldProperties(name="my_geopoint", type="geopoint")
    ]

    assert_raises_errors(lambda: check_data(data, resource_properties), ValueError, 1)


def test_rejects_incorrect_resource_properties():
    """Should throw an error if the resource properties are incorrect."""
    assert_raises_check_errors(lambda: check_data(example_data(), ResourceProperties()))



from typing import Any, Callable

from pytest import raises

from seedcase_sprout.check_datapackage.check_error import CheckError


def assert_raises_errors(
    fn: Callable[[], Any],
    error_type: type[BaseException],
    error_count: int | None = None,
) -> None:
    """Asserts that the function raises a group of errors of the given type."""
    with raises(ExceptionGroup) as error_info:
        fn()

    errors = error_info.value.exceptions
    assert all(isinstance(error, error_type) for error in errors)
    if error_count is not None:
        assert len(errors) == error_count


def assert_raises_check_errors(
    fn: Callable[[], Any], error_count: int | None = None
) -> None:
    """Asserts that the function raises a group of `CheckError`s."""
    assert_raises_errors(fn, CheckError, error_count)



from pytest import mark

from seedcase_sprout import (
    PackageProperties,
    as_readme_text,
    example_package_properties,
    example_resource_properties_all_types,
)


def test_creates_readme():
    """Should create a README with basic information about the properties."""
    properties = example_package_properties()
    resource = example_resource_properties_all_types()
    assert resource.schema
    assert resource.schema.fields
    properties.resources = [resource]

    readme = as_readme_text(properties)

    assert str(resource.title) in readme
    assert all([str(field.name) in readme for field in resource.schema.fields])


def test_creates_readme_with_empty_values():
    """Should be able to create a README for an empty set of properties."""
    assert as_readme_text(PackageProperties())


@mark.parametrize(
    "description, expected",
    [
        # Empty or whitespace-only.
        ("", ""),
        ("    ", ""),
        # Single line text.
        ("Non-indented one-line text.", "Non-indented one-line text."),
        ("    Indented one-line text.", "Indented one-line text."),
        (
            "    Indented one-line text with trailing whitespace.   ",
            "Indented one-line text with trailing whitespace.   ",
        ),
        # Multiline text.
        ("Non-indented\nmultiline\ntext.", "Non-indented\nmultiline\ntext."),
        ("    Indented\n    multiline text.", "Indented\nmultiline text."),
        (
            """Non-indented first line,
            indented second line.""",
            "Non-indented first line,\nindented second line.",
        ),
        (
            """
            Indented multiline
            text
            """,
            "Indented multiline\ntext",
        ),
        # Tab indentation.
        (
            "\tIndented with tab\n\tAlso indented with tab",
            "Indented with tab\nAlso indented with tab",
        ),
        # Mixed indentation.
        (
            "\tIndented with tab\n    Indented with spaces\n\t  Mixed indented line",
            "Indented with tab\nIndented with spaces\nMixed indented line",
        ),
        (
            "  Indented with 2 spaces\n    Indented with 4 spaces\n  Back to 2 spaces",
            "Indented with 2 spaces\nIndented with 4 spaces\nBack to 2 spaces",
        ),
        # Multi-level list (relative indentation removed, not ideal).
        (
            """
            Description with multilevel list:
            - Item 1
                - Item 2""",
            "Description with multilevel list:\n- Item 1\n- Item 2",
        ),
    ],
)
def test_dedents_multiline_description(description, expected):
    """Should be able to dedent description."""
    properties = PackageProperties(
        name="diabetes-hypertension-study",
        description=description,
    )

    readme_text = as_readme_text(properties)
    assert expected in readme_text



from typing import Any

from seedcase_sprout.check_datapackage import RequiredFieldType


def get_blank_value_for_type(type: RequiredFieldType) -> str | list[Any] | None:
    """Returns the blank value for each type of (required) field.

    Args:
        type: The type of the field.

    Returns:
        The corresponding blank value.
    """
    match type:
        case RequiredFieldType.str:
            return ""
        case RequiredFieldType.list:
            return []
        case _:
            return None



from datetime import datetime
from typing import Any, TypeVar

T = TypeVar("T")


def _get_iso_timestamp() -> str:
    """Gets the current ISO timestamp compliant with the Data Package spec.

    Returns:
        The current ISO timestamp as a string. E.g. `2024-05-14T05:00:01+00:00`.
    """
    return datetime.now().astimezone().isoformat(timespec="seconds")


def _get_nested_attr(
    base_object: Any, attributes: str, default: T | None = None
) -> T | None:
    """Returns the attribute specified by `attributes`.

    Tries to resolve the chain of attributes against `base_object`. Returns None
    or the specified default value if the attribute chain cannot be resolved.

    Args:
        base_object: The object to start resolving the attributes from.
        attributes: The chain of attributes as a dot-separated string.
        default: The default value to return if the attributes cannot be resolved.
            Defaults to None.

    Returns:
        The value at the end of the attribute chain.

    Raises:
        ValueError: If the attribute chain contains an element that is not a valid
            identifier.

    Examples:
        ```{python}
        class Inner:
            pass
        class Middle:
            inner: Inner = Inner()
        class Outer:
            middle: Middle = Middle()

        _get_nested_attr(Outer(), "middle.inner")
        ```
    """
    attributes_list = attributes.split(".")
    if any(not attribute.isidentifier() for attribute in attributes_list):
        raise ValueError(
            "`attributes` should contain valid identifiers separated by `.`."
        )

    try:
        for attribute in attributes_list:
            base_object = getattr(base_object, attribute)
    except AttributeError:
        return default

    return default if base_object is None else base_object



from typing import Any

from seedcase_sprout.check_datapackage import CheckError
from seedcase_sprout.sprout_checks.check_fields_present import (
    check_fields_present,
)
from seedcase_sprout.sprout_checks.check_required_package_properties_not_blank import (  # noqa: E501
    check_required_package_properties_not_blank,
)
from seedcase_sprout.sprout_checks.required_fields import (
    PACKAGE_SPROUT_REQUIRED_FIELDS,
)


def get_sprout_package_errors(properties: dict[str, Any]) -> list[CheckError]:
    """Checks the package `properties` against Sprout-specific requirements only.

    Args:
        properties: The package properties.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    errors = check_required_package_properties_not_blank(properties)
    errors += check_fields_present(properties, PACKAGE_SPROUT_REQUIRED_FIELDS)
    return errors



from datetime import datetime
from pathlib import Path
from uuid import uuid4

import polars as pl

from seedcase_sprout.check_data import check_data
from seedcase_sprout.check_properties import (
    check_resource_properties,
)
from seedcase_sprout.constants import BATCH_TIMESTAMP_FORMAT
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import ResourceProperties


def write_resource_batch(
    data: pl.DataFrame,
    resource_properties: ResourceProperties,
    package_path: Path | None = None,
) -> Path:
    """Write the tidied data into the resource's batch data folder.

    Use this function to write the original data that is in a tidy Polars
    DataFrame to the resource folder provided by the `path` property in the
    `resource_properties`. The function saves the DataFrame as file with a
    timestamped, unique name, as a backup. If the data is exactly the same as
    the data in one of the existing batch files, it won't be duplicated. See the
    [design](https://sprout.seedcase-project.org/docs/design/) docs for an
    explanation of this batch file. Data is always checked against the
    `resource_properties` before it is written to the batch folder.

    Args:
        data: A Polars DataFrame object with the data to write to the batch folder.
        resource_properties: The properties object for the specific resource.
            Use `read_properties()` to read the properties for the resource.
        package_path: The path to the data package root folder (where `datapackage.json`
            is located). Defaults to the current working directory.

    Returns:
        The path to the written Parquet resource file.

    Raises:
        ExceptionGroup[CheckError]: If the resource properties are incorrect.
        ValueError: If column names in the data are incorrect.
        ExceptionGroup[ValueError]: If data types in the data are incorrect.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            resource_properties = sp.read_properties().resources[0]
            sp.write_resource_batch(
                data=sp.example_data(),
                resource_properties=resource_properties,
            )
        ```
    """
    check_resource_properties(resource_properties)
    check_data(data, resource_properties)

    package = PackagePath(package_path)
    resource_name = str(resource_properties.name)

    for batch_path in package.resource_batch_files(resource_name):
        batch = pl.read_parquet(batch_path)
        if data.equals(batch):
            return batch_path

    batch_path = package.resource_batch(resource_name)
    batch_path.mkdir(exist_ok=True, parents=True)
    # TODO: Move out some of this into the create_batch_file_name during refactoring
    batch_file_path = batch_path / _create_batch_file_name()

    data.write_parquet(batch_file_path)

    return batch_file_path


def _create_batch_file_name() -> Path:
    return Path(f"{_get_compact_iso_timestamp()}-{uuid4()}.parquet")


def _get_compact_iso_timestamp() -> str:
    """Gets the current timestamp in a compact ISO format.

    Returns:
        The current compact ISO timestamp as a string in the format defined by
        BATCH_TIMESTAMP_FORMAT.
    """
    return datetime.now().strftime(BATCH_TIMESTAMP_FORMAT)



import os
import re

import polars as pl
from polars.testing import assert_frame_equal
from pytest import raises

from seedcase_sprout.constants import BATCH_TIMESTAMP_PATTERN
from seedcase_sprout.examples import (
    example_data,
    example_resource_properties,
)
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.write_resource_batch import write_resource_batch


def test_writes_correct_resource_batch_file(tmp_path):
    """Writes tidy resource batch file correctly."""
    # Given
    (tmp_path / "resources" / example_resource_properties().name).mkdir(parents=True)

    # When
    batch_path = write_resource_batch(
        example_data(), example_resource_properties(), tmp_path
    )
    batch_data = pl.read_parquet(batch_path)

    # Then
    assert batch_path.exists()
    assert f"resources/{example_resource_properties().name}/batch" in str(batch_path)
    assert len(re.findall(BATCH_TIMESTAMP_PATTERN, batch_path.stem)) == 1
    assert_frame_equal(batch_data, example_data(), check_exact=True)


def test_does_not_duplicate_batch_file(tmp_cwd):
    """Should not create a new batch file when the data has already been saved in an
    existing batch file."""
    # Given
    data = example_data()
    resource_properties = example_resource_properties()
    existing_batch_path = write_resource_batch(data, resource_properties)

    # When
    new_batch_path = write_resource_batch(data, resource_properties)

    # Then
    assert new_batch_path == existing_batch_path
    assert len(PackagePath().resource_batch_files(str(resource_properties.name))) == 1
    assert_frame_equal(pl.read_parquet(new_batch_path), data)


def test_creates_new_batch_file_when_data_not_duplicate(tmp_cwd):
    """Should create a new batch file when the data is different from data in
    existing batch files."""
    # Given
    existing_data = example_data()
    new_data = example_data()
    new_data[2, "name"] = "Bob"
    resource_properties = example_resource_properties()
    existing_batch_path = write_resource_batch(existing_data, resource_properties)

    # When
    new_batch_path = write_resource_batch(new_data, resource_properties)

    # Then
    assert_frame_equal(pl.read_parquet(existing_batch_path), existing_data)
    assert_frame_equal(pl.read_parquet(new_batch_path), new_data)


def test_writes_correct_resource_batch_file_with_unordered_columns(tmp_path):
    """Writes batch file correctly even if columns aren't in the order expected by the
    resource properties."""
    # Given
    (tmp_path / "resources" / example_resource_properties().name).mkdir(parents=True)
    data = example_data().select(["value", "name", "id"])

    # When
    batch_path = write_resource_batch(data, example_resource_properties(), tmp_path)
    batch_data = pl.read_parquet(batch_path)

    # Then
    assert batch_path.exists()
    assert_frame_equal(batch_data, data, check_exact=True)


def test_throws_error_if_data_do_not_match_example_resource_properties(tmp_path):
    """Throws ExceptionGroup if data don't match resource properties."""
    with raises(ValueError):
        write_resource_batch(pl.DataFrame(), example_resource_properties(), tmp_path)


def test_throws_error_if_data_do_not_match_resource_properties(tmp_path):
    """Throws ExceptionGroup if data don't match resource properties (extra column)."""
    # Given
    os.chdir(tmp_path)
    (tmp_path / "resources" / example_resource_properties().name).mkdir(parents=True)

    data = example_data().insert_column(2, pl.Series("extra_column", [1, 2, 3]))

    # When
    with raises(ValueError):
        write_resource_batch(data, example_resource_properties(), tmp_path)


def test_uses_cwd_if_no_path_provided(tmp_cwd):
    """If no path is provided, should use the cwd as the package root."""
    # Given
    package_path = PackagePath(tmp_cwd)
    resource_properties = example_resource_properties()
    assert resource_properties.name
    package_path.resource(resource_properties.name).mkdir(parents=True)

    # When
    batch_file = write_resource_batch(example_data(), resource_properties)

    # Then
    assert batch_file.exists()
    assert package_path.resource_batch(resource_properties.name) in batch_file.parents



from enum import Enum
from importlib.resources import files
from pathlib import Path


# Data Package standard required fields and their types
class RequiredFieldType(str, Enum):
    """A class enumerating allowed types for required fields."""

    str = "str"
    list = "list"
    any = "any"


COMPLEX_VALIDATORS = {"allOf", "anyOf", "oneOf"}

DATA_PACKAGE_SCHEMA_PATH = Path(
    str(
        files("seedcase_sprout.check_datapackage.schemas").joinpath(
            "data-package-schema.json"
        )
    )
)

NAME_PATTERN = r"^[a-z0-9._-]+$"

PACKAGE_RECOMMENDED_FIELDS = {
    "name": RequiredFieldType.str,
    "id": RequiredFieldType.str,
    "licenses": RequiredFieldType.list,
}

PACKAGE_REQUIRED_FIELDS = {
    "resources": RequiredFieldType.list,
}

RESOURCE_REQUIRED_FIELDS = {
    "name": RequiredFieldType.str,
    "path": RequiredFieldType.str,
    "data": RequiredFieldType.any,
}

# From https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
SEMVER_PATTERN = (
    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0"
    r"|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>"
    r"[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)



from pytest import mark

from seedcase_sprout.sprout_checks.check_no_inline_data import check_no_inline_data
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)


@mark.parametrize("index", [None, 2])
def test_passes_if_data_not_set(index):
    """Should pass if inline data is not set."""
    assert check_no_inline_data({}, index) == []


@mark.parametrize("index", [None, 2])
def test_error_found_if_data_is_set(index):
    """Should find an error if inline data is set."""
    properties = {"data": "some data"}

    errors = check_no_inline_data(properties, index)

    assert len(errors) == 1
    assert errors[0].message
    assert errors[0].json_path == get_json_path_to_resource_field("data", index)
    assert errors[0].validator == "inline-data"



import json
from pathlib import Path
from typing import Any

from seedcase_sprout.write_file import write_file


def _write_json(json_object: dict[str, Any], path: Path) -> Path:
    """Writes an object as an indented JSON string to the specified location.

    Args:
        json_object: The object to write to the file. Must be JSON serialisable.
        path: The path to the file with name and extension.

    Returns:
        The path to the newly created JSON file.

    Raises:
        FileNotFoundError: If the parent folder of the file doesn't exist.
        TypeError: If the object is not JSON serialisable.
    """
    return write_file(json.dumps(json_object, indent=2, ensure_ascii=False), path)



import tempfile
from pathlib import Path

from pytest import raises

from seedcase_sprout.check_data import check_data
from seedcase_sprout.check_properties import (
    check_package_properties,
    check_properties,
    check_resource_properties,
)
from seedcase_sprout.examples import (
    ExamplePackage,
    example_data,
    example_data_all_types,
    example_package_properties,
    example_resource_properties,
    example_resource_properties_all_types,
)
from seedcase_sprout.read_properties import read_properties


def test_creates_correct_package_properties_object():
    assert check_package_properties(example_package_properties())


def test_creates_correct_resource_properties_object():
    assert check_resource_properties(example_resource_properties())


def test_creates_correct_resource_properties_object_all_types():
    assert check_resource_properties(example_resource_properties_all_types())


def test_example_data_matches_resource_properties():
    data = example_data()
    assert check_data(data, example_resource_properties()) is data


def test_example_data_matches_resource_properties_all_types():
    data = example_data_all_types()
    assert check_data(data, example_resource_properties_all_types()) is data


def test_creates_package_with_resources():
    """Should create: package folder, datapackage.json, README."""
    with ExamplePackage() as package_path:
        # Test root folder
        assert package_path.root().is_dir()
        assert len(list(package_path.root().iterdir())) == 3
        # Test README
        assert package_path.readme().is_file()
        assert package_path.readme().read_text()
        # Test datapackage.json
        assert package_path.properties().is_file()
        properties = read_properties(package_path.properties())
        assert properties.resources
        assert check_resource_properties(properties.resources[0])
        # Test resources
        assert properties.resources and len(properties.resources) == 1


def test_creates_package_without_resources():
    """Should create: package folder, datapackage.json, README."""
    with ExamplePackage(with_resources=False) as package_path:
        # Test root folder
        assert package_path.root().is_dir()
        assert len(list(package_path.root().iterdir())) == 2
        # Test README
        assert package_path.readme().is_file()
        assert package_path.readme().read_text()
        # Test datapackage.json
        assert package_path.properties().is_file()
        properties = read_properties(package_path.properties())
        assert check_properties(properties)
        assert properties.name == package_path.root().stem
        # Test no resources
        assert not properties.resources


def test_changes_cwd_to_package_root_in_package_context():
    """Within the context, the cwd should be the package root.
    After exiting the context, the cwd should be reset to the original location."""
    original_cwd = Path.cwd()
    with ExamplePackage() as package_path:
        # Resolve paths to ensure symlinks are compared correctly on MacOS as well
        # (i.e., `/var` -> `/private/var` on MacOS)
        assert Path.cwd().resolve() == package_path.root().resolve()

    assert Path.cwd() == original_cwd


def test_restores_cwd_when_error_raised_in_context():
    """The original cwd should be restored after exiting the context, even if an error
    was raised inside the context."""
    original_cwd = Path.cwd()
    with ExamplePackage(), raises(ValueError):
        raise ValueError("An error!")

    assert Path.cwd() == original_cwd


def test_manages_temp_folder_correctly():
    """The package root should be in a temporary folder.
    After exiting the context, the package root should not exist."""
    temp_root = Path(tempfile.gettempdir())
    with ExamplePackage() as package_path:
        assert temp_root in package_path.root().parents

    assert not package_path.root().exists()


def test_can_use_context_without_referencing_package_path_explicitly():
    """It should be possible to omit the package path from function calls within the
    context."""
    with ExamplePackage():
        properties = read_properties()
        assert check_properties(properties)


def test_can_set_name_for_package():
    """It should be possible to specify a custom name for the package."""
    package_name = "basking-shark-study"
    with ExamplePackage(package_name=package_name) as package_path:
        assert read_properties().name == package_name
        assert package_path.root().stem == package_name



from pathlib import Path

from pytest import fixture, mark

from seedcase_sprout.properties import ResourceProperties
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)
from seedcase_sprout.sprout_checks.get_sprout_resource_errors import (
    get_sprout_resource_errors,
)
from seedcase_sprout.sprout_checks.required_fields import (
    RESOURCE_SPROUT_REQUIRED_FIELDS,
)


@fixture
def properties():
    return ResourceProperties(
        name="resource-1",
        title="Resource 1",
        description="A resource.",
    ).compact_dict


def test_passes_full_resource_properties(properties):
    """Should pass with a full set of resource properties."""
    assert get_sprout_resource_errors(properties) == []


@mark.parametrize("index", [None, 2])
def test_error_found_if_inline_data_is_set(properties, index):
    """Should find an error if inline data is set."""
    properties["data"] = "some data"

    errors = get_sprout_resource_errors(properties, index=index)

    assert len(errors) == 1
    assert errors[0].json_path == get_json_path_to_resource_field("data", index)
    assert errors[0].validator == "inline-data"


@mark.parametrize("index", [None, 2])
@mark.parametrize("name,type", RESOURCE_SPROUT_REQUIRED_FIELDS.items())
def test_error_found_if_fields_are_blank(properties, name, type, index):
    """Should find an error if there is one required field that is present but blank."""
    properties[name] = get_blank_value_for_type(type)

    errors = get_sprout_resource_errors(properties, index=index)

    assert len(errors) == 1
    assert errors[0].json_path == get_json_path_to_resource_field(name, index)
    assert errors[0].validator == "blank"


@mark.parametrize("index", [None, 2])
@mark.parametrize("name", RESOURCE_SPROUT_REQUIRED_FIELDS.keys())
def test_error_found_if_required_fields_are_missing(properties, name, index):
    """Should find an error if there is a missing required field."""
    del properties[name]

    errors = get_sprout_resource_errors(properties, index=index)

    assert len(errors) == 1
    assert errors[0].json_path == get_json_path_to_resource_field(name, index)
    assert errors[0].validator == "required"


@mark.parametrize("path", ["", [], str(Path("resources", "1"))])
def test_error_found_if_data_path_is_incorrect_(properties, path):
    """Should find one error if `path` is not a string or has the wrong format."""
    properties["path"] = path

    errors = get_sprout_resource_errors(properties)

    assert len(errors) == 1
    assert errors[0].json_path == get_json_path_to_resource_field("path")


def test_ignores_path_if_name_incorrect(properties):
    """Should not check the path if the name is incorrect."""
    properties["name"] = "name with spaces"
    properties["path"] = "bad/path"

    assert get_sprout_resource_errors(properties) == []



from pytest import mark

from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.check_error_matcher import CheckErrorMatcher


@mark.parametrize(
    "message,matcher,expected",
    [
        ("", "", True),
        ("Match!", None, True),
        ("Match!", "", True),
        ("Complete match", "Complete match", True),
        ("Beginning matches", "Beginning", True),
        ("End matches", "d matches", True),
        ("", "No match", False),
        ("No match", "something else", False),
    ],
)
def test_matches_message(message, matcher, expected):
    """Should match if the matcher's message is a substring of the error's message."""
    assert (
        CheckErrorMatcher(message=matcher).matches(
            CheckError(message=message, json_path="$.name", validator="")
        )
        is expected
    )


@mark.parametrize(
    "json_path,matcher,expected",
    [
        ("$.match", None, True),
        ("$.match", "", True),
        ("", "", True),
        ("$.same", r"\$\.same", True),
        ("$.start.match", r"\$\.start", True),
        ("$.start.match", r"^\$\.start", True),
        ("$.match.end", "end", True),
        ("$.match.end", r"end$", True),
        ("$.match.middle.match", "middle", True),
        ("$.match.list[999].match", r"list\[\d+\]", True),
        ("$.not.last.field", r"\.last$", False),
        ("$.not.first.field", r"^\$\.first", False),
        ("$.no.match", "other", False),
    ],
)
def test_matches_json_path(json_path, matcher, expected):
    """Should match if the regular expression in the matcher's `json_path` matches the
    `json_path` of the error."""
    assert (
        CheckErrorMatcher(json_path=matcher).matches(
            CheckError(message="Hello", json_path=json_path, validator="")
        )
        is expected
    )


@mark.parametrize(
    "validator,matcher,expected",
    [
        ("match", None, True),
        ("match", "match", True),
        ("", "", True),
        ("no-match", "match", False),
        ("no-match", "", False),
        ("", "no-match", False),
    ],
)
def test_matches_validator(validator, matcher, expected):
    """Should match if the matcher's validator is the same as the error's."""
    assert (
        CheckErrorMatcher(validator=matcher).matches(
            CheckError(message="Hello", json_path="", validator=validator)
        )
        is expected
    )


@mark.parametrize(
    "message,json_path,validator,expected",
    [
        ("name' is a", "name", "required", True),
        ("name' is a", "name", "no-match", False),
        ("name' is a", r"no\.match", "required", False),
        ("no match", "name", "required", False),
        ("no match", "name", "no-match", False),
        ("no match", r"no\.match", "no-match", False),
    ],
)
def test_matches_on_all_fields(message, json_path, validator, expected):
    """Should only match if all fields match."""
    assert (
        CheckErrorMatcher(
            message=message, json_path=json_path, validator=validator
        ).matches(
            CheckError(
                message="'name' is a required property",
                json_path="$.name",
                validator="required",
            )
        )
        is expected
    )



from typing import Any

from seedcase_sprout.check_datapackage import CheckError, RequiredFieldType
from seedcase_sprout.sprout_checks.check_fields_not_blank import (
    SPROUT_BLANK_ERROR_MESSAGE,
)
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)


def check_list_item_field_not_blank(
    properties: dict[str, Any],
    list_name: str,
    field_name: str,
    field_type: RequiredFieldType = RequiredFieldType.str,
) -> list[CheckError]:
    """Checks that the specified field of items in a list is not blank.

    Args:
        properties: The properties object containing the list.
        list_name: The name of the list field.
        field_name: The name of the item field.
        field_type: The type of the item field. Defaults to str.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    return [
        CheckError(
            message=SPROUT_BLANK_ERROR_MESSAGE.format(field_name=field_name),
            json_path=f"$.{list_name}[{index}].{field_name}",
            validator="blank",
        )
        for index, item in enumerate(properties.get(list_name, []))
        if item.get(field_name) == get_blank_value_for_type(field_type)
    ]



from pathlib import Path

from pytest import fixture, raises

from seedcase_sprout.examples import (
    example_package_properties,
    example_resource_properties,
)
from seedcase_sprout.internals import _read_json, _write_json
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import (
    PackageProperties,
)
from seedcase_sprout.write_properties import write_properties


@fixture
def properties() -> PackageProperties:
    properties = example_package_properties()
    properties.resources = [example_resource_properties()]
    return properties


@fixture
def path(tmp_path: Path) -> Path:
    return tmp_path / "datapackage.json"


def assert_file_contains(path: Path, expected_properties: PackageProperties) -> None:
    new_properties = _read_json(path)
    assert new_properties == expected_properties.compact_dict


def test_writes_properties_when_file_does_not_exist(path, properties):
    """Should write properties to file when the file doesn't yet exist."""
    assert write_properties(properties, path) == path
    assert_file_contains(path, properties)


def test_overwrites_properties_when_file_exists(path, properties):
    """Should overwrite properties when the file already exists"""
    old_properties = properties.compact_dict | {
        "name": "old-name",
        "resources": [{"name": "old-resource-name"}, {"name": "another-old-resource"}],
    }
    _write_json(old_properties, path)

    assert write_properties(properties, path) == path
    assert_file_contains(path, properties)


def test_writes_properties_using_utf8(path, properties):
    """Should write properties to file using UTF-8."""
    properties.description = (
        "Håkan Ørsted's favourite letters: æ ø å Æ Ø Å é μῆνιν ἄειδε θεὰ"
    )

    write_properties(properties, path)

    assert properties.description in path.read_bytes().decode("utf-8")


def test_throws_error_if_error_in_package_properties(path, properties):
    """Should throw `CheckError`s if there are errors in the package properties."""
    properties.id = None

    with raises(ExceptionGroup):
        write_properties(properties, path)


def test_throws_error_if_error_in_resource_properties(path, properties):
    """Should throw `CheckError`s if there are errors in the resource properties."""
    properties.resources[0].name = "invalid name with spaces"

    with raises(ExceptionGroup):
        write_properties(properties, path)


def test_writes_properties_to_cwd_if_no_path_provided(tmp_cwd, properties):
    """If no path is provided, should use datapackage.json in the cwd."""
    assert write_properties(properties) == PackagePath(tmp_cwd).properties()


def test_writes_properties_with_dedented_descriptions(path, properties):
    """Should write properties to file with dedented description."""
    # Simple test here bc it's tested thoroughly in the `as_readme_text` tests.
    indented_text = """
        Indented description with leading spaces.
        \t Multiline text with tab and space.
        """

    properties.description = indented_text
    properties.resources[0].description = indented_text
    properties.resources[0].schema.fields[0].description = indented_text

    write_properties(properties, path)

    dedented_text = (
        "Indented description with leading spaces.\\nMultiline text with tab and space."
    )
    file_text = path.read_text()
    assert file_text.count(dedented_text) == 3


def test_throws_error_if_resource_description_is_none(path, properties):
    """Should throw an error if the required resource description is None, i.e.,
    dedentation works and the subsequent check fails."""
    properties.resources[0].description = None

    with raises(ExceptionGroup):
        write_properties(properties, path)


def test_writes_properties_when_field_description_is_none(path, properties):
    """Should write properties to file without optional field description."""
    properties.resources[0].schema.fields[0].description = None
    write_properties(properties, path)

    assert_file_contains(path, properties)



"""Internal functions for the package."""

from .check import _check_is_dir, _check_is_file
from .create import (
    _create_resource_data_path,
    _create_resource_properties_script_filename,
)
from .functionals import _map, _map2
from .get import _get_iso_timestamp, _get_nested_attr
from .read import _read_json
from .to import _to_camel_case, _to_dedented, _to_snake_case
from .write import _write_json

__all__ = [
    "_check_is_file",
    "_check_is_dir",
    "_create_resource_data_path",
    "_create_resource_properties_script_filename",
    "_to_snake_case",
    "_get_iso_timestamp",
    "_get_nested_attr",
    "_map",
    "_map2",
    "_read_json",
    "_write_json",
    "_to_camel_case",
    "_to_dedented",
]



from typing import Any

from pytest import fixture, mark

from seedcase_sprout.examples import example_package_properties
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)
from seedcase_sprout.sprout_checks.get_sprout_package_errors import (
    get_sprout_package_errors,
)
from seedcase_sprout.sprout_checks.required_fields import (
    PACKAGE_SPROUT_REQUIRED_FIELDS,
)


@fixture
def properties() -> dict[str, Any]:
    return example_package_properties().compact_dict


def test_passes_full_package_properties(properties):
    """Should pass with a full set of package properties."""
    assert get_sprout_package_errors(properties) == []


@mark.parametrize("name,type", PACKAGE_SPROUT_REQUIRED_FIELDS.items())
def test_error_found_if_fields_are_blank(properties, name, type):
    """Should find an error if there is one required field that is present but blank."""
    properties[name] = get_blank_value_for_type(type)

    errors = get_sprout_package_errors(properties)

    assert len(errors) == 1
    assert errors[0].json_path == f"$.{name}"
    assert errors[0].validator == "blank"


@mark.parametrize("name", PACKAGE_SPROUT_REQUIRED_FIELDS.keys())
def test_error_found_if_required_fields_are_missing(properties, name):
    """Should find an error if there is a missing required field."""
    del properties[name]

    errors = get_sprout_package_errors(properties)

    assert len(errors) == 1
    assert errors[0].json_path == f"$.{name}"
    assert errors[0].validator == "required"



from typing import cast

import polars as pl
from polars.testing import assert_frame_equal

from seedcase_sprout.examples import (
    ExamplePackage,
    example_data,
    example_data_all_types,
    example_resource_properties,
    example_resource_properties_all_types,
)
from seedcase_sprout.properties import ResourceProperties
from seedcase_sprout.read_properties import read_properties
from seedcase_sprout.write_resource_data import write_resource_data
from tests.assert_raises_errors import (
    assert_raises_check_errors,
    assert_raises_errors,
)


def test_writes_data_to_correct_location():
    """Should write data to correct location with custom path and default path."""
    with ExamplePackage() as package_path:
        resources = cast(list[ResourceProperties], read_properties().resources)
        resource_properties = resources[0]
        data = example_data()
        # Resolve paths to ensure symlinks are compared correctly on MacOS as well
        # (i.e., `/var` -> `/private/var` on MacOS)
        assert resource_properties.name
        expected_path = package_path.resource_data(resource_properties.name).resolve()

        # With custom path
        data_path = write_resource_data(data, resource_properties, package_path.root())

        assert data_path.resolve() == expected_path
        assert_frame_equal(pl.read_parquet(data_path), data)

        # With default path
        data_path = write_resource_data(data, resource_properties)

        assert data_path.resolve() == expected_path
        assert_frame_equal(pl.read_parquet(data_path), data)


def test_rewrites_data_file():
    """New data files should rewrite old data files."""
    with ExamplePackage():
        old_resources = cast(list[ResourceProperties], read_properties().resources)
        old_resource_properties = old_resources[0]
        old_data = example_data()
        new_resource_properties = example_resource_properties_all_types()
        new_resource_properties.name = old_resource_properties.name
        new_resource_properties.path = old_resource_properties.path
        new_data = example_data_all_types()
        write_resource_data(old_data, old_resource_properties)

        data_path = write_resource_data(new_data, new_resource_properties)

        assert_frame_equal(pl.read_parquet(data_path), new_data)


def test_throws_error_if_resource_properties_incorrect():
    """Should throw an error if the resource properties are incorrect."""
    resource_properties = example_resource_properties()
    resource_properties.name = "spaces in name"

    assert_raises_check_errors(
        lambda: write_resource_data(example_data(), resource_properties)
    )


def test_throws_error_if_properties_do_not_match_data():
    """Should throw an error if the resource properties and the data don't match."""
    resource_properties = example_resource_properties()
    assert resource_properties.schema and resource_properties.schema.fields
    resource_properties.schema.fields[0].type = "yearmonth"

    assert_raises_errors(
        lambda: write_resource_data(example_data(), resource_properties), ValueError
    )
